#//-----------------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------

from .uvm_object_globals import (UVM_DEFAULT_POLICY, UVM_INFO, UVM_LOW,
    UVM_NORADIX, UVM_REFERENCE)
from .uvm_object import UVMObject
from .uvm_scope_stack import UVMScopeStack
from .uvm_pool import UVMPool
from .sv import sv



class UVMComparer:
    """
    The UVMComparer class provides a policy object for doing comparisons. The
    policies determine how miscompares are treated and counted. Results of a
    comparison are stored in the comparer object. The <uvm_object::compare>
    and <uvm_object::do_compare> methods are passed a UVMComparer policy
    object.

    """

    def __init__(self):
        #
        #  // Variable: policy
        #  //
        #  // Determines whether comparison is UVM_DEEP, UVM_REFERENCE, or UVM_SHALLOW.
        #
        #  uvm_recursion_policy_enum policy = UVM_DEFAULT_POLICY
        self.policy = UVM_DEFAULT_POLICY


        #  // Variable: show_max
        #  //
        #  // Sets the maximum number of messages to send to the printer for miscompares
        #  // of an object.
        #  int unsigned show_max = 1
        self.show_max = 1


        #  // Variable: verbosity
        #  //
        #  // Sets the verbosity for printed messages.
        #  //
        #  // The verbosity setting is used by the messaging mechanism to determine
        #  // whether messages should be suppressed or shown.
        #  int unsigned verbosity = UVM_LOW
        self.verbosity = UVM_LOW


        #  // Variable: sev
        #  //
        #  // Sets the severity for printed messages.
        #  //
        #  // The severity setting is used by the messaging mechanism for printing and
        #  // filtering messages.
        #  uvm_severity sev = UVM_INFO
        self.sev = UVM_INFO


        #  // Variable: miscompares
        #  //
        #  // This string is reset to an empty string when a comparison is started.
        #  //
        #  // The string holds the last set of miscompares that occurred during a
        #  // comparison.
        #  string miscompares = ""
        self.miscompares = ""


        #  // Variable: physical
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The abstract and physical settings allow an object to distinguish between
        #  // two different classes of fields.
        #  //
        #  // It is up to you, in the <uvm_object::do_compare> method, to test the
        #  // setting of this field if you want to use the physical trait as a filter.
        #  bit physical = 1
        self.physical = 1


        #  // Variable: abstract
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The abstract and physical settings allow an object to distinguish between
        #  // two different classes of fields.
        #  //
        #  // It is up to you, in the <uvm_object::do_compare> method, to test the
        #  // setting of this field if you want to use the abstract trait as a filter.
        #  bit abstract = 1
        self.abstract = 1


        #  // Variable: check_type
        #  //
        #  // This bit determines whether the type, given by <uvm_object::get_type_name>,
        #  // is used to verify that the types of two objects are the same.
        #  //
        #  // This bit is used by the <compare_object> method. In some cases it is useful
        #  // to set this to 0 when the two operands are related by inheritance but are
        #  // different types.
        #  bit check_type = 1
        self.check_type = 1

        #
        #  // Variable: result
        #  //
        #  // This bit stores the number of miscompares for a given compare operation.
        #  // You can use the result to determine the number of miscompares that
        #  // were found.
        #  int unsigned result = 0
        self.result = 0

        self.depth = 0  # current depth of objects
        self.compare_map = UVMPool()  # uvm_object [uvm_object]
        self.scope = UVMScopeStack()


    def compare_field(self, name, lhs, rhs, size, radix=UVM_NORADIX):
        """
          Function: compare_field

          Compares two integral values.

          The `name` input is used for purposes of storing and printing a miscompare.

          The left-hand-side `lhs` and right-hand-side `rhs` objects are the two
          objects used for comparison.

          The size variable indicates the number of bits to compare; size must be
          less than or equal to 4096.

          The radix is used for reporting purposes, the default radix is hex.

         virtual function bit compare_field (string name,
                                             uvm_bitstream_t lhs,
                                             uvm_bitstream_t rhs,
                                             int size,
                                             uvm_radix_enum radix=UVM_NORADIX);
        Args:
            name:
            lhs:
            rhs:
            size:
            radix:
        Returns:
        """
        mask = 0
        msg = ""
        if size <= 64:
            return self.compare_field_int(name, lhs, rhs, size, radix)
        #
        #    mask = -1
        #    mask >>= (UVM_STREAMBITS-size)
        #    if((lhs & mask) !== (rhs & mask)):
        #      UVMObject._m_uvm_status_container.scope.set_arg(name)
        #      case (radix)
        #        UVM_BIN: begin
        #              $swrite(msg, "lhs = 'b%0b : rhs = 'b%0b",
        #                       lhs&mask, rhs&mask)
        #             end
        #        UVM_OCT: begin
        #              $swrite(msg, "lhs = 'o%0o : rhs = 'o%0o",
        #                       lhs&mask, rhs&mask)
        #             end
        #        UVM_DEC: begin
        #              $swrite(msg, "lhs = %0d : rhs = %0d",
        #                       lhs&mask, rhs&mask)
        #             end
        #        UVM_TIME: begin
        #            $swrite(msg, "lhs = %0t : rhs = %0t",
        #               lhs&mask, rhs&mask)
        #        end
        #        UVM_STRING: begin
        #              $swrite(msg, "lhs = %0s : rhs = %0s",
        #                       lhs&mask, rhs&mask)
        #             end
        #        UVM_ENUM: begin
        #              //Printed as decimal, user should cuse compare string for enum val
        #              $swrite(msg, "lhs = %0d : rhs = %0d",
        #                       lhs&mask, rhs&mask)
        #              end
        #        default: begin
        #              $swrite(msg, "lhs = 'h%0x : rhs = 'h%0x",
        #                       lhs&mask, rhs&mask)
        #             end
        #      endcase
        #      print_msg(msg)
        #      return 0
        #    end
        #    return 1
        #  endfunction


    def compare_field_int(self, name, lhs, rhs, size, radix=UVM_NORADIX):
        """
          Function: compare_field_int

          This method is the same as `compare_field` except that the arguments are
          small integers, less than or equal to 64 bits. It is automatically called
          by `compare_field` if the operand size is less than or equal to 64.

         virtual function bit compare_field_int (string name,
                                                 uvm_integral_t lhs,
                                                 uvm_integral_t rhs,
                                                 int size,
                                                 uvm_radix_enum radix=UVM_NORADIX);
        Args:
            name:
            lhs:
            rhs:
            size:
            radix:
        Returns:
        """
        mask = 0x0
        msg = ""
        mask = -1
        mask >>= (64-size)
        if (lhs & mask) != (rhs & mask):
            UVMObject._m_uvm_status_container.scope.set_arg(name)
            msg = "lhs = {} : rhs = {}".format(lhs, rhs)
            # TODO pretty formatting
            #      case (radix)
            #        UVM_BIN: begin
            #              $swrite(msg, "lhs = 'b%0b : rhs = 'b%0b",
            #                       lhs&mask, rhs&mask)
            #             end
            #        UVM_OCT: begin
            #              $swrite(msg, "lhs = 'o%0o : rhs = 'o%0o",
            #                       lhs&mask, rhs&mask)
            #             end
            #        UVM_DEC: begin
            #              $swrite(msg, "lhs = %0d : rhs = %0d",
            #                       lhs&mask, rhs&mask)
            #             end
            #        UVM_TIME: begin
            #            $swrite(msg, "lhs = %0t : rhs = %0t",
            #               lhs&mask, rhs&mask)
            #        end
            #        UVM_STRING: begin
            #              $swrite(msg, "lhs = %0s : rhs = %0s",
            #                       lhs&mask, rhs&mask)
            #             end
            #        UVM_ENUM: begin
            #              //Printed as decimal, user should cuse compare string for enum val
            #              $swrite(msg, "lhs = %0d : rhs = %0d",
            #                       lhs&mask, rhs&mask)
            #              end
            #        default: begin
            #              $swrite(msg, "lhs = 'h%0x : rhs = 'h%0x",
            #                       lhs&mask, rhs&mask)
            #             end
            #      endcase
            self.print_msg(msg)
            return 0
            #    end
        return 1
        #  endfunction


    def compare_field_real(self, name, lhs, rhs):
        """
        This method is the same as <compare_field> except that the arguments are
        real numbers.

        Args:
            lhs: First float to compare
            rhs: Second float to compare
        Returns:
            bool: True if numbers match, False otherwise
        """
        if lhs != rhs:
            UVMObject._m_uvm_status_container.scope.set_arg(name)
            msg = "lhs = " + str(lhs) + " : rhs = " + str(rhs)
            self.print_msg(msg)
            return False
        return True


    def compare_object(self, name, lhs, rhs):
        """
        Compares two class objects using the `policy` knob to determine whether the
        comparison should be deep, shallow, or reference.

        The name input is used for purposes of storing and printing a miscompare.

        The `lhs` and `rhs` objects are the two objects used for comparison.

        The `check_type` determines whether or not to verify the object
        types match (the return from ~lhs.get_type_name()~ matches
        ~rhs.get_type_name()~).

        Args:
            name (str): Extra info for printing miscompare.
            lhs (UVMObject): First object to compare.
            rhs (UVMObject): Second object to compare.
        Returns:
            bool: True if objects match, False otherwise.
        """
        if rhs == lhs:
            return True

        if self.policy == UVM_REFERENCE and lhs != rhs:
            UVMObject._m_uvm_status_container.scope.set_arg(name)
            self.print_msg_object(lhs, rhs)
            return False

        if rhs is None or lhs is None:
            UVMObject._m_uvm_status_container.scope.set_arg(name)
            self.print_msg_object(lhs, rhs)
            return False  # miscompare

        UVMObject._m_uvm_status_container.scope.down(name)
        res = lhs.compare(rhs, self)
        UVMObject._m_uvm_status_container.scope.up()
        return res


    def compare_string(self, name, lhs, rhs):
        """
        Compares two string variables.

        The `name` input is used for purposes of storing and printing a miscompare.
        The `lhs` and `rhs` objects are the two objects used for comparison.

        Args:
            name (str): Extra info for printing miscompare.
            lhs (str): First str to compare.
            rhs (str): Second str to compare.
        Returns:
            bool: True if strings match, False otherwise.
        """
        msg = ""
        if lhs != rhs:
            UVMObject._m_uvm_status_container.scope.set_arg(name)
            msg = "lhs = \"" + lhs + "\" : rhs = \"" + rhs + "\""
            self.print_msg(msg)
            return False
        return True



    def print_msg(self, msg):
        """
        Causes the error count to be incremented and the message, `msg`, to be
        appended to the `miscompares` string (a newline is used to separate
        messages).

        If the message count is less than the `show_max` setting, then the message
        is printed to standard-out using the current verbosity and severity
        settings. See the `verbosity` and `sev` variables for more information.

        Args:
            msg (str): Message to print
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        root = cs.get_root()
        #
        self.result += 1
        if self.result <= self.show_max:
            msg = ("Miscompare for " +
                UVMObject._m_uvm_status_container.scope.get() + ": " + msg)
            root.uvm_report(self.sev, "MISCMP", msg, self.verbosity, "`uvm_file", "`uvm_line")
        self.miscompares = (self.miscompares + UVMObject._m_uvm_status_container.scope.get()
            + ": " + msg + "\n")


    def print_rollup(self, rhs, lhs):
        """
        Internal methods - do not call directly

          print_rollup
          ------------

         Need this function because sformat doesn't support objects
        Args:
            rhs:
            lhs:
        """
        pass
        # TODO
        #    uvm_root root
        #    uvm_coreservice_t cs
        #
        #    string msg
        #    cs = uvm_coreservice_t::get()
        #    root = cs.get_root()
        #    if(UVMObject._m_uvm_status_container.scope.depth() == 0):
        #      if(result && (show_max || (uvm_severity'(sev) != UVM_INFO))):
        #        if(show_max < result)
        #           $swrite(msg, "%0d Miscompare(s) (%0d shown) for object ",
        #             result, show_max)
        #        else begin
        #           $swrite(msg, "%0d Miscompare(s) for object ", result)
        #        end
        #
        #        root.uvm_report(sev, "MISCMP", $sformatf("%s%s@%0d vs. %s@%0d", msg,
        #                        lhs.get_name(), lhs.get_inst_id(), rhs.get_name(),
        #                        rhs.get_inst_id()),
        #			verbosity, `uvm_file, `uvm_line)
        #      end
        #    end
        #  endfunction


    def print_msg_object(self, lhs, rhs):
        """

        Args:
            lhs (UVMObject):
            rhs (UVMObject):
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        root = cs.get_root()

        self.result += 1
        if (self.result <= self.show_max):
            lhs_id = 0
            if lhs is not None:
                lhs_id = lhs.get_inst_id()
            rhs_id = 0
            if rhs is not None:
                rhs_id = rhs.get_inst_id()
            root.uvm_report(self.sev, "MISCMP",
                sv.sformatf("Miscompare for %0s: lhs = @%0d : rhs = @%0d",
                UVMObject._m_uvm_status_container.scope.get(), lhs_id, rhs_id,
                self.verbosity, "`uvm_file", "`uvm_line"))

            self.miscompares = sv.sformatf("%s%s: lhs = @%0d : rhs = @%0d",
                self.miscompares, UVMObject._m_uvm_status_container.scope.get(),
                lhs_id, rhs_id)


    #  // init ??
    #
    #  static function UVMComparer init()
    #    if(uvm_default_comparer==null) uvm_default_comparer=new
    #    return uvm_default_comparer
    #  endfunction



#//------------------------------------------------------------------------------
#// Copyright 2010-2011 Synopsys, Inc.
#// Copyright 2010-2011 Cadence Design Systems, Inc.
#// Copyright 2019 Tuomas Poikela (tpoikela)
#// All Rights Reserved Worldwide
#//
#// Licensed under the Apache License, Version 2.0 (the "License"); you may
#// not use this file except in compliance with the License.  You may obtain
#// a copy of the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#// Unless required by applicable law or agreed to in writing, software
#// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
#// License for the specific language governing permissions and limitations
#// under the License.
#//------------------------------------------------------------------------------

import cocotb
from cocotb.triggers import RisingEdge, Timer
from cocotb.binary import BinaryValue

from uvm.base import *
from uvm.reg import UVMRegAdapter
from uvm.reg.uvm_reg_model import *
from uvm.macros import *
from uvm.comps import *
from uvm.seq import *

class reg_rw(UVMSequenceItem):
    #
    #   rand bit          read
    #   rand bit   [31:0] addr
    #   rand logic [31:0] data
    #   rand bit   [3:0] byte_en
    #
    #   `uvm_object_utils_begin(reg_rw)
    #     `uvm_field_int(read, UVM_ALL_ON | UVM_NOPACK)
    #     `uvm_field_int(addr, UVM_ALL_ON | UVM_NOPACK)
    #     `uvm_field_int(data, UVM_ALL_ON | UVM_NOPACK)
    #     `uvm_field_int(byte_en, UVM_ALL_ON | UVM_NOPACK)
    #   `uvm_object_utils_end
    #
    def __init__(self, name="reg_rw"):
        UVMSequenceItem.__init__(self, name)
        r32b = range(0, (1 << 32) - 1)
        self.read = False
        self.rand('read', [False, True])
        self.addr = 0
        self.rand('addr', r32b)
        self.data = 0
        self.rand('data', r32b)
        self.byte_en = 0x0
        self.rand('byte_en', list(range(0, 16)))

    def convert2string(self):
        tt = "WRITE"
        if self.read:
            tt = "READ"
        return sv.sformatf("%s addr=%0h data=%0h be=%b",
                tt,self.addr,self.data,self.byte_en)
    #endclass: reg_rw
uvm_object_utils(reg_rw)


class reg_sequencer(UVMSequencer):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)


uvm_component_utils(reg_sequencer)


class reg_monitor(UVMMonitor):

    def __init__(self, name, parent=None):
        UVMMonitor.__init__(self, name, parent)
        self.ap = UVMAnalysisPort("ap", self)


class reg_driver(UVMComponent):


    def __init__(self, name, parent=None):
        UVMComponent.__init__(self, name, parent)
        self.m_parent = parent
        self.seqr_port = UVMSeqItemPullPort("seqr_port",self)
        self.T = reg_rw
        self.dut = None
        self.set_nbits(32)

    def set_nbits(self, nbits):
        self.n_bits = nbits
        self.bit_en = 0x0
        for n in range(self.n_bits):
            self.bit_en <<= 1
            self.bit_en |= 0x1

    async def run_phase(self, phase):
        mon = self.m_parent.get_child("mon")
        while True:
            rw = []  # reg_rw
            await self.seqr_port.peek(rw)  # aka 'get_next_rw'
            rw = rw[0]
            await self.drive_transaction(rw)
            mon.ap.write(rw)
            rw = []
            await self.seqr_port.get(rw)  # aka 'item_done'


    async def drive_transaction(self, rw):
        if rw.read is False:
            #wdata = BinaryValue(value=rw.data,n_bits=self.n_bits)
            await RisingEdge(self.dut.clk)
            await Timer(0)
            self.dut.we <= 1
            #print('DDWW data is ' + str(wdata))
            #self.dut.data_in <= wdata
            self.dut.data_in <= rw.data & self.bit_en
            self.dut.addr_in <= rw.addr

            for i in range(2):
                await RisingEdge(self.dut.clk)
            await Timer(0)
            self.dut.we <= 0
            uvm_info("REG_DRIVER WRITE", "Wrote value to DUT: " + str(rw.data) +
                ' addr:' + str(rw.addr), UVM_LOW)
        else:
            await RisingEdge(self.dut.clk)
            self.dut.addr_in <= rw.addr
            self.dut.read <= 0x1
            await RisingEdge(self.dut.clk)
            rw.data = self.dut.data_out.value.integer
            await RisingEdge(self.dut.clk)
            self.dut.read <= 0x0
            uvm_info("REG_DRIVER READ", "Read value from DUT: " + str(rw.data) +
                ' addr:' + str(rw.addr), UVM_LOW)


uvm_component_utils(reg_driver)


class reg_agent(UVMAgent):


    def __init__(self, name, parent=None):
        UVMAgent.__init__(self, name, parent)
        self.sqr = reg_sequencer("sqr", self)
        self.drv = reg_driver("drv", self)
        self.mon = reg_monitor("mon", self)
        self.nbits = 32


    def set_nbits(self, n):
        self.nbits = n

    def build_phase(self, phase):
        UVMAgent.build_phase(self, phase)
        dut = []
        if not UVMConfigDb.get(self, "", "dut", dut):
            uvm_fatal("REG_AGENT", "No 'dut' found inf config DB")
        self.drv.dut = dut[0]

    def connect_phase(self, phase):
        self.drv.seqr_port.connect(self.sqr.seq_item_export)
        self.drv.set_nbits(self.nbits)


uvm_component_utils(reg_agent)


class reg2rw_adapter(UVMRegAdapter):

    def __init__(self, name="reg2rw_adapter"):
        UVMRegAdapter.__init__(self, name)
        self.supports_byte_enable = True

    def reg2bus(self, rw):
        bus = reg_rw.type_id.create("rw")
        bus.read    = (rw.kind == UVM_READ)
        bus.addr    = rw.addr
        bus.data    = rw.data
        bus.byte_en = rw.byte_en
        return bus

    def bus2reg(self, bus_item, rw):
        bus = bus_item
        rw.kind    = UVM_WRITE

        if bus.read:
            rw.kind = UVM_READ
        rw.addr    = bus.addr
        rw.data    = bus.data
        rw.byte_en = bus.byte_en
        rw.status  = UVM_IS_OK
        return rw


uvm_object_utils(reg2rw_adapter)



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2014 NVIDIA Corporation
#//   Copyright 2019-2021 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from typing import List

from .uvm_sequencer_param_base import UVMSequencerParamBase
from ..tlm1.uvm_sqr_connections import UVMSeqItemPullImp
from ..macros import uvm_component_utils
from ..base.uvm_globals import (uvm_check_output_args, uvm_zero_delay,
    uvm_report_info, uvm_report_error)
from ..base.uvm_object_globals import *

FATAL_MSG1 = ("Item_done() called with no outstanding requests." +
    " Each call to item_done() must be paired with a previous call to "
    + " get_next_item().")


class UVMSequencer(UVMSequencerParamBase):
    """
    Group: Sequencer Interface
    This is an interface for communicating with sequencers.

    The interface is defined as::

      Requests:
       async def get_next_item      (request: List)
       async def try_next_item      (request: List)
       async def get                (request: List)
       async def peek               (request: List)
      Responses:
       def item_done          (response=None)
       async def put          (response: UVMSequenceItem)
      Sync Control:
       async def wait_for_sequences()
       def has_do_available()

    See `UVMSqrIfBase` for information about this interface.
    """

    def __init__(self, name, parent=None):
        """
        Standard component constructor that creates an instance of this class
        using the given `name` and `parent`, if any.

        Args:
            name (str): Name of the sequencer
            parent (UVMComponent): Parent component
        """
        UVMSequencerParamBase.__init__(self, name, parent)
        #  // Variable: seq_item_export
        #  // This export provides access to this sequencer's implementation of the
        #  // sequencer interface.
        self.seq_item_export = UVMSeqItemPullImp("seq_item_export", self)
        self.sequence_item_requested = False
        self.get_next_item_called = False

    def stop_sequences(self):
        """
        Tells the sequencer to kill all sequences and child sequences currently
        operating on the sequencer, and remove all requests, locks and responses
        that are currently queued.  This essentially resets the sequencer to an
        idle state.
        """
        super().stop_sequences()
        self.sequence_item_requested  = 0
        self.get_next_item_called     = 0
        # Empty the request fifo
        if self.m_req_fifo.used():
            uvm_report_info(self.get_full_name(),
                "Sequences stopped. Removing request from sequencer fifo")
            t = []
            while self.m_req_fifo.try_get(t):
                t = []

    #  extern virtual function string get_type_name()



    async def get_next_item(self, t):
        """
        Retrieves the next available item from a sequence.

        Args:
            t (list): Empty list into which item is appended
        """
        uvm_check_output_args([t])
        # req_item = None

        # If a sequence_item has already been requested, then get_next_item()
        # should not be called again until item_done() has been called.

        if self.get_next_item_called is True:
            self.uvm_report_error(self.get_full_name(),
                "Get_next_item called twice without item_done or get in between", UVM_NONE)

        if self.sequence_item_requested is False:
            await self.m_select_sequence()

        # Set flag indicating that the item has been requested to ensure that item_done or get
        # is called between requests
        self.sequence_item_requested = True
        self.get_next_item_called = True
        await self.m_req_fifo.peek(t)

    async def try_next_item(self, t: List):
        """
        Retrieves the next available item from a sequence if one is available.

        Args:
            t (List): Empty list into which item is appended
        """
        if self.get_next_item_called == 1:
            uvm_report_error(self.get_full_name(),
                "get_next_item/try_next_item called twice without item_done or get in between",
                UVM_NONE)
            return

        # allow state from last transaction to settle such that sequences'
        # relevancy can be determined with up-to-date information
        await self.wait_for_sequences()

        # choose the sequence based on relevancy
        selected_sequence = self.m_choose_next_request()

        # return if none available
        if selected_sequence == -1:
            # t = None
            return

        # now, allow chosen sequence to resume
        self.m_set_arbitration_completed(self.arb_sequence_q[selected_sequence].request_id)
        seq = self.arb_sequence_q[selected_sequence].sequence_ptr
        self.arb_sequence_q.delete(selected_sequence)
        self.m_update_lists()
        self.sequence_item_requested = True
        self.get_next_item_called = 1

        # give it one NBA to put a new item in the fifo
        await self.wait_for_sequences()

        # attempt to get the item; if it fails, produce an error and return
        if not self.m_req_fifo.try_peek(t):
            uvm_report_error("TRY_NEXT_BLOCKED", ("try_next_item: the selected sequence '" +
                seq.get_full_name() + "' did not produce an item within an NBA delay. " +
                "Sequences should not consume time between calls to start_item and finish_item. " +
                "Returning null item."), UVM_NONE)

    def item_done(self, item=None):
        """
        Indicates that the request is completed.

        Args:
            item (UVMSequenceItem): Related sequence item.
        """
        t = []
        # Set flag to allow next get_next_item or peek to get a new sequence_item
        self.sequence_item_requested = False
        self.get_next_item_called = False

        if self.m_req_fifo.try_get(t) is False:
            self.uvm_report_fatal(self.get_full_name(), FATAL_MSG1)
        else:
            t = t[0]
            self.m_wait_for_item_sequence_id = t.get_sequence_id()
            self.m_wait_for_item_transaction_id = t.get_transaction_id()

        if item is not None:
            self.seq_item_export.put_response(item)

        # Grant any locks as soon as possible
        self.grant_queued_locks()


    async def put(self, t):
        """
        Sends a response back to the sequence that issued the request.

        Args:
            t (UVMSequenceItem): Response item.
        """
        self.put_response(t)
        await uvm_zero_delay()


    async def get(self, t: List):
        """
        Retrieves the next available item from a sequence into the given list.

        Args:
            t (List): List to hold the response
        """
        if self.sequence_item_requested == 0:
            await self.m_select_sequence()
        self.sequence_item_requested = 1
        await self.m_req_fifo.peek(t)
        self.item_done()


    async def peek(self, t: List):
        """
        Gets the current request item if one is in the FIFO.

        Args:
            t (List): List for the output request.
        """
        if self.sequence_item_requested == 0:
            await self.m_select_sequence()

        # Set flag indicating that the item has been requested to ensure that item_done or get
        # is called between requests
        self.sequence_item_requested = 1
        await self.m_req_fifo.peek(t)


    """
    Documented here for clarity, implemented in `UVMSequencerBase`

    Task: `wait_for_sequences`
    Waits for a sequence to have a new item available.


    Function: `has_do_available`
    Returns 1 if any sequence running on this sequencer is ready to supply
    a transaction, 0 otherwise.

    """

    #
    #  //-----------------
    #  // Internal Methods
    #  //-----------------
    #  // Do not use directly, not part of standard
    #
    #  extern function void         item_done_trigger(RSP item = null)
    #  function RSP                 item_done_get_trigger_data()
    #    return last_rsp(0)
    #  endfunction
    #  extern protected virtual function int m_find_number_driver_connections()


uvm_component_utils(UVMSequencer)


#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#function string uvm_sequencer::get_type_name()
#  return "uvm_sequencer"
#endfunction
#
#//-----------------
#// Internal Methods
#//-----------------
#
#// m_find_number_driver_connections
#// --------------------------------
#// Counting the number of of connections is done at end of
#// elaboration and the start of run.  If the user neglects to
#// call super in one or the other, the sequencer will still
#// have the correct value
#
#function int uvm_sequencer::m_find_number_driver_connections()
#  uvm_port_component_base provided_to_port_list[string]
#  uvm_port_component_base seq_port_base
#
#  // Check that the seq_item_pull_port is connected
#  seq_port_base = seq_item_export.get_comp()
#  seq_port_base.get_provided_to(provided_to_port_list)
#  return provided_to_port_list.num()
#endfunction
#
#
#
#


#// item_done_trigger
#// -----------------
#
#function void uvm_sequencer::item_done_trigger(RSP item = null)
#  item_done(item)
#endfunction



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer

from uvm.comps.uvm_test import UVMTest
from uvm.macros import *
from uvm import (run_test, uvm_debug, UVMDebug, UVMPhase, UVMReportCatcher,
    THROW, UVM_FATAL, UVM_INFO, UVMReportCb, UVMCallbacksBase)

from tb_timer import *
from tb_env import tb_env

UVMDebug.DEBUG = False
UVMPhase.m_phase_trace = False
UVMCallbacksBase.m_tracing = True


class timeout_catcher(UVMReportCatcher):
    def __init__(self, name="timeout_catcher"):
        super().__init__(name)
        self.num_fatals = 0

    def catch(self):
        uvm_info("CATCHER_ACTIVE", "Catcher was activated", UVM_LOW)
        if self.get_severity() == UVM_FATAL and self.get_id() == "TIMEOUT":
            self.set_severity(UVM_INFO)
            self.num_fatals += 1
        return THROW


class test_comp(UVMTest):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.timer = tb_timer("global_timer", None)
        self.catcher = timeout_catcher()
        UVMReportCb.add(None, self.catcher)

    
    async def pre_main_phase(self, phase):
        phase.raise_objection(self)
        await Timer(100, "NS")
        phase.drop_objection(self)

    
    async def main_phase(self, phase):
        phase.raise_objection(self)
        await Timer(100, "NS")
        # Will cause a time-out
        # because we forgot to drop the objection
        phase.drop_objection(self)


    
    async def shutdown_phase(self, phase):
        phase.raise_objection(self)
        await Timer(100, "NS")
        phase.drop_objection(self)


uvm_component_utils(test_comp)


@cocotb.test()
async def initial_begin(dut):
    env = tb_env("env")
    await run_test("test_comp")



#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
"""
Title: Sequence-Related Functions

Group: Sequence Action Functions

NOTE: Each macro defines `seq_obj` argument, which is `self` argument within
the sequence member functions. `seq_obj` is not present in SystemVerilog UVM
since the context comes from where the macro is used.

These functions are used to start sequences and sequence items on the default
sequencer, `m_sequencer`. This is determined a number of ways.
- the sequencer handle provided in the `UVMSequenceBase.start` method
- the sequencer used by the parent sequence
- the sequencer that was set using the `UVMSequenceItem::set_sequencer` method

"""

import cocotb
from .uvm_message_defines import uvm_warning, uvm_fatal


def uvm_create(seq_obj, SEQ_OR_ITEM, m_sequencer):
    """
    This action creates the item or sequence using the factory.  It intentionally
    does zero processing.  After this action completes, the user can manually set
    values, manipulate rand_mode and constraint_mode, etc.

    Args:
        seq_obj (UVMSequence): Sequence from which the function is called
    """
    return uvm_create_on(seq_obj, SEQ_OR_ITEM, m_sequencer)




async def uvm_do(seq_obj, SEQ_OR_ITEM):
    """
    This macro takes as an argument a uvm_sequence_item variable or object.
    The argument is created using `uvm_create` if necessary,
    then randomized.
    In the case of an item, it is randomized after the call to
    `UVMSequenceBase.start_item()` returns.
    This is called late-randomization.
    In the case of a sequence, the sub-sequence is started using
    `UVMSequenceBase.start_item()` with `call_pre_post` set to 0.
    In the case of an item,
    the item is sent to the driver through the associated sequencer.
    
    For a sequence item, the following are called, in order::
    
       uvm_create(item)
       sequencer.wait_for_grant(prior) (task)
       this.pre_do(1)                  (task)
       item.randomize()
       this.mid_do(item)               (func)
       sequencer.send_request(item)    (func)
       sequencer.wait_for_item_done()  (task)
       this.post_do(item)              (func)
    
    For a sequence, the following are called, in order::
    
       uvm_create(sub_seq)
       sub_seq.randomize()
       sub_seq.pre_start()         (task)
       this.pre_do(0)              (task)
       this.mid_do(sub_seq)        (func)
       await sub_seq.body()              (task)
       this.post_do(sub_seq)       (func)
       sub_seq.post_start()        (task)
   """
    
    await uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, seq_obj.m_sequencer, -1, [])


#// MACRO: `uvm_do_pri
#//
#//| `uvm_do_pri(SEQ_OR_ITEM, PRIORITY)
#//
#// This is the same as `uvm_do except that the sequence item or sequence is
#// executed with the priority specified in the argument
#
#`define uvm_do_pri(SEQ_OR_ITEM, PRIORITY) \
#  `uvm_do_on_pri_with(SEQ_OR_ITEM, m_sequencer, PRIORITY, {})


async def uvm_do_with(seq_obj, SEQ_OR_ITEM, *CONSTRAINTS):
    """
    This is the same as `uvm_do except that the constraint block in the 2nd
    argument is applied to the item or sequence in a randomize with statement
    before execution.
    """
    await uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, seq_obj.m_sequencer, -1,
            *CONSTRAINTS)


async def uvm_do_pri_with(seq_obj, SEQ_OR_ITEM, PRIORITY, *CONSTRAINTS):
    """
    This is the same as `uvm_do_pri except that the given constraint block is
    applied to the item or sequence in a randomize with statement before
    execution.
    """
    await uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, seq_obj.m_sequencer,
            PRIORITY, *CONSTRAINTS)


"""
Group: Sequence on Sequencer Action Macros

These macros are used to start sequences and sequence items on a specific
sequencer. The sequence or item is created and executed on the given
sequencer.
"""


def uvm_create_on(seq_obj, SEQ_OR_ITEM, SEQR):
    """
    This is the same as <`uvm_create> except that it also sets the parent sequence
    to the sequence in which the macro is invoked, and it sets the sequencer to
    the specified ~SEQR~ argument.
    """
    w_ = SEQ_OR_ITEM.get_type()
    if w_ is None:
        uvm_fatal("uvm_create_on",
            "Object {} not registered with factory".format(SEQ_OR_ITEM.get_name()))
        return None
    else:
        return seq_obj.create_item(w_, SEQR, "SEQ_OR_ITEM")


async def uvm_do_on(seq_obj, SEQ_OR_ITEM, SEQR):
    """
    This is the same as <`uvm_do> except that it also sets the parent sequence to
    the sequence in which the macro is invoked, and it sets the sequencer to the
    specified ~SEQR~ argument.
    """
    await uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, SEQR, -1, {})


#// MACRO: `uvm_do_on_pri
#//
#//| `uvm_do_on_pri(SEQ_OR_ITEM, SEQR, PRIORITY)
#//
#// This is the same as <`uvm_do_pri> except that it also sets the parent sequence
#// to the sequence in which the macro is invoked, and it sets the sequencer to
#// the specified ~SEQR~ argument.
#
#`define uvm_do_on_pri(SEQ_OR_ITEM, SEQR, PRIORITY) \
#  `uvm_do_on_pri_with(SEQ_OR_ITEM, SEQR, PRIORITY, {})


async def uvm_do_on_with(seq_obj, SEQ_OR_ITEM, SEQR, *CONSTRAINTS):
    """
    This is the same as `uvm_do_with` except that it also sets the parent
    sequence to the sequence in which the macro is invoked, and it sets the
    sequencer to the specified ~SEQR~ argument.
    The user must supply the constraints using lambdas.

    An example call::

        await uvm_do_on_with(self, sys0_seq, None,
            lambda num_blk_seq: num_blk_seq == 10,
            lambda blk_level_delay_ns: blk_level_delay_ns in [10, 20],
        )

    Note that variables used in lambdas must exist, or an exception is thrown
    due to randomization error.
    """
    await uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, SEQR, -1,
            *CONSTRAINTS)

#`define uvm_do_on_pri_with(SEQ_OR_ITEM, SEQR, PRIORITY, CONSTRAINTS) \
#  begin \
#  uvm_sequence_base __seq; \
#  `uvm_create_on(SEQ_OR_ITEM, SEQR) \
#  if (!$cast(__seq,SEQ_OR_ITEM)) start_item(SEQ_OR_ITEM, PRIORITY);\
#  if ((__seq == null || !__seq.do_not_randomize) && !SEQ_OR_ITEM.randomize() with CONSTRAINTS ) begin \
#    `uvm_warning("RNDFLD", "Randomization failed in uvm_do_with action") \
#  end\
#  if (!$cast(__seq,SEQ_OR_ITEM)) finish_item(SEQ_OR_ITEM, PRIORITY); \
#  else __seq.start(SEQR, this, PRIORITY, 0); \
#  end
async def uvm_do_on_pri_with(seq_obj, SEQ_OR_ITEM, SEQR, PRIORITY, *CONSTRAINTS):
    """
    This is the same as uvm_do_pri_with except that it also sets the parent
    sequence to the sequence in which the function is invoked, and it sets the
    sequencer to the specified ~SEQR~ argument.

    Args:
        seq_obj (UVMSequence): Sequence context.
        SEQ_OR_ITEM (UVMSequence|UVMSequenceItem): 
        SEQR (UVMSequencer): Runs sequence on this sequencer.
        CONSTRAINTS (constraints): Randomization constraints
    """
    from ..seq.uvm_sequence import UVMSequence
    _seq = uvm_create_on(seq_obj, SEQ_OR_ITEM, SEQR)
    if isinstance(_seq, UVMSequence):
        if SEQ_OR_ITEM.do_not_randomize == 0:
            if SEQ_OR_ITEM.randomize_with(*CONSTRAINTS) is False:
                uvm_warning("RNDFLD", "Randomization failed in uvm_do_with action")
        await SEQ_OR_ITEM.start(SEQR, seq_obj, PRIORITY, 0)
    else:
        # TODO handle constraints
        await seq_obj.start_item(SEQ_OR_ITEM, PRIORITY)
        if SEQ_OR_ITEM.randomize_with(*CONSTRAINTS) is False:
            uvm_warning("RNDFLD", "Randomization failed in uvm_do_with action")
        await seq_obj.finish_item(SEQ_OR_ITEM, PRIORITY)


"""
Group: Sequence Action Functions for Pre-Existing Sequences

These functions are used to start sequences and sequence items that do not
need to be created.
"""

#`define uvm_send(SEQ_OR_ITEM) \
#  `uvm_send_pri(SEQ_OR_ITEM, -1)
async def uvm_send(seq_obj, SEQ_OR_ITEM):
    """
    This function processes the item or sequence that has been created using
    `uvm_create`.  The processing is done without randomization.  Essentially, an
    `uvm_do` without the create or randomization.

    Args:
        seq_obj (UVMSequence): Sequence context.
        SEQ_OR_ITEM (UVMSequence|UVMSequenceItem): 
    """
    await uvm_send_pri(seq_obj, SEQ_OR_ITEM, -1)



#`define uvm_send_pri(SEQ_OR_ITEM, PRIORITY) \
#  begin \
#  uvm_sequence_base __seq; \
#  if (!$cast(__seq,SEQ_OR_ITEM)) begin \
#     start_item(SEQ_OR_ITEM, PRIORITY);\
#     finish_item(SEQ_OR_ITEM, PRIORITY);\
#  end \
#  else __seq.start(__seq.get_sequencer(), this, PRIORITY, 0);\
#  end
async def uvm_send_pri(seq_obj, SEQ_OR_ITEM, PRIORITY):
    """
    This is the same as `uvm_send` except that the sequence item or sequence is
    executed with the priority specified in the argument.

    Args:
        seq_obj (UVMSequence): Sequence context.
        SEQ_OR_ITEM (UVMSequence|UVMSequenceItem): 
        PRIORITY (int): Priority of the sequence
    """
    from ..seq.uvm_sequence import UVMSequence
    _seq = SEQ_OR_ITEM
    if isinstance(_seq, UVMSequence):
        await _seq.start(_seq.get_sequencer(), seq_obj, PRIORITY, 0)
    else:
        await seq_obj.start_item(SEQ_OR_ITEM, PRIORITY)
        await seq_obj.finish_item(SEQ_OR_ITEM, PRIORITY)


#// MACRO: `uvm_rand_send
#//
#//| `uvm_rand_send(SEQ_OR_ITEM)
#//
#// This macro processes the item or sequence that has been already been
#// allocated (possibly with `uvm_create). The processing is done with
#// randomization.  Essentially, an `uvm_do without the create.
#
#`define uvm_rand_send(SEQ_OR_ITEM) \
#  `uvm_rand_send_pri_with(SEQ_OR_ITEM, -1, {})


#// MACRO: `uvm_rand_send_pri
#//
#//| `uvm_rand_send_pri(SEQ_OR_ITEM, PRIORITY)
#//
#// This is the same as `uvm_rand_send except that the sequence item or sequence
#// is executed with the priority specified in the argument.
#
#`define uvm_rand_send_pri(SEQ_OR_ITEM, PRIORITY) \
#  `uvm_rand_send_pri_with(SEQ_OR_ITEM, PRIORITY, {})


#// MACRO: `uvm_rand_send_with
#//
#//| `uvm_rand_send_with(SEQ_OR_ITEM, CONSTRAINTS)
#//
#// This is the same as `uvm_rand_send except that the given constraint block is
#// applied to the item or sequence in a randomize with statement before
#// execution.
#
#`define uvm_rand_send_with(SEQ_OR_ITEM, CONSTRAINTS) \
#  `uvm_rand_send_pri_with(SEQ_OR_ITEM, -1, CONSTRAINTS)


#// MACRO: `uvm_rand_send_pri_with
#//
#//| `uvm_rand_send_pri_with(SEQ_OR_ITEM, PRIORITY, CONSTRAINTS)
#//
#// This is the same as `uvm_rand_send_pri except that the given constraint block
#// is applied to the item or sequence in a randomize with statement before
#// execution.
#
#`define uvm_rand_send_pri_with(SEQ_OR_ITEM, PRIORITY, CONSTRAINTS) \
#  begin \
#  uvm_sequence_base __seq; \
#  if (!$cast(__seq,SEQ_OR_ITEM)) start_item(SEQ_OR_ITEM, PRIORITY);\
#  else __seq.set_item_context(this,SEQ_OR_ITEM.get_sequencer()); \
#  if ((__seq == null || !__seq.do_not_randomize) && !SEQ_OR_ITEM.randomize() with CONSTRAINTS ) begin \
#    `uvm_warning("RNDFLD", "Randomization failed in uvm_rand_send_with action") \
#  end\
#  if (!$cast(__seq,SEQ_OR_ITEM)) finish_item(SEQ_OR_ITEM, PRIORITY);\
#  else __seq.start(__seq.get_sequencer(), this, PRIORITY, 0);\
#  end


#`define uvm_create_seq(UVM_SEQ, SEQR_CONS_IF) \
#  `uvm_create_on(UVM_SEQ, SEQR_CONS_IF.consumer_seqr) \

#`define uvm_do_seq(UVM_SEQ, SEQR_CONS_IF) \
#  `uvm_do_on(UVM_SEQ, SEQR_CONS_IF.consumer_seqr) \

#`define uvm_do_seq_with(UVM_SEQ, SEQR_CONS_IF, CONSTRAINTS) \
#  `uvm_do_on_with(UVM_SEQ, SEQR_CONS_IF.consumer_seqr, CONSTRAINTS) \


#//-----------------------------------------------------------------------------
#//
#// Group- Sequence Library
#//
#//-----------------------------------------------------------------------------


#// MACRO: `uvm_add_to_sequence_library
#//
#// Adds the given sequence ~TYPE~ to the given sequence library ~LIBTYPE~
#//
#//| `uvm_add_to_seq_lib(TYPE,LIBTYPE)
#//
#// Invoke any number of times within a sequence declaration to statically add
#// that sequence to one or more sequence library types. The sequence will then
#// be available for selection and execution in all instances of the given
#// sequencer types.
#//
#//| class seqA extends uvm_sequence_base #(simple_item);
#//|
#//|    function new(string name=`"TYPE`");
#//|      super.new(name);
#//|    endfunction
#//|
#//|    `uvm_object_utils(seqA)
#//|
#//|    `uvm_add_to_seq_lib(seqA, simple_seq_lib_RST)
#//|    `uvm_add_to_seq_lib(seqA, simple_seq_lib_CFG)
#//|
#//|    virtual task body(); \
#//|      `uvm_info("SEQ_START", {"Executing sequence '", get_full_name(),
#//|                             "' (",get_type_name(),")"},UVM_HIGH)
#//|      #10;
#//|    endtask
#//|
#//|  endclass
#
#
#`define uvm_add_to_seq_lib(TYPE,LIBTYPE) \
#   static bit add_``TYPE``_to_seq_lib_``LIBTYPE =\
#      LIBTYPE::m_add_typewide_sequence(TYPE::get_type());



#// MACRO: `uvm_sequence_library_utils
#//
#//| `uvm_sequence_library_utils(TYPE)
#//
#// Declares the infrastructure needed to define extensions to the
#// <uvm_sequence_library> class. You define new sequence library subtypes
#// to statically specify sequence membership from within sequence
#// definitions. See also <`uvm_add_to_sequence_library> for more information.
#//
#//
#//| typedef simple_seq_lib uvm_sequence_library #(simple_item);
#//|
#//| class simple_seq_lib_RST extends simple_seq_lib;
#//|
#//|   `uvm_object_utils(simple_seq_lib_RST)
#//|
#//|   `uvm_sequence_library_utils(simple_seq_lib_RST)
#//|
#//|   function new(string name="");
#//|     super.new(name);
#//|   endfunction
#//|
#//| endclass
#//
#// Each library, itself a sequence, can then be started independently
#// on different sequencers or in different phases of the same sequencer.
#// See <uvm_sequencer_base::start_phase_sequence> for information on
#// starting default sequences.
#
#`define uvm_sequence_library_utils(TYPE) \
#\
#   static protected uvm_object_wrapper m_typewide_sequences[$]; \
#   \
#   function void init_sequence_library(); \
#     foreach (TYPE::m_typewide_sequences[i]) \
#       sequences.push_back(TYPE::m_typewide_sequences[i]); \
#   endfunction \
#   \
#   static function void add_typewide_sequence(uvm_object_wrapper seq_type); \
#     if (m_static_check(seq_type)) \
#       TYPE::m_typewide_sequences.push_back(seq_type); \
#   endfunction \
#   \
#   static function void add_typewide_sequences(uvm_object_wrapper seq_types[$]); \
#     foreach (seq_types[i]) \
#       TYPE::add_typewide_sequence(seq_types[i]); \
#   endfunction \
#   \
#   static function bit m_add_typewide_sequence(uvm_object_wrapper seq_type); \
#     TYPE::add_typewide_sequence(seq_type); \
#     return 1; \
#   endfunction



#//-----------------------------------------------------------------------------
#//
#// Group: Sequencer Subtypes
#//
#//-----------------------------------------------------------------------------
#
#
#// MACRO: `uvm_declare_p_sequencer
#//
#// This macro is used to declare a variable ~p_sequencer~ whose type is
#// specified by ~SEQUENCER~.
#//
#//| `uvm_declare_p_sequencer(SEQUENCER)
#//
#// The example below shows using the `uvm_declare_p_sequencer macro
#// along with the uvm_object_utils macros to set up the sequence but
#// not register the sequence in the sequencer's library.
#//
#//| class mysequence extends uvm_sequence#(mydata);
#//|   `uvm_object_utils(mysequence)
#//|   `uvm_declare_p_sequencer(some_seqr_type)
#//|   task body;
#//|     //Access some variable in the user's custom sequencer
#//|     if(p_sequencer.some_variable) begin
#//|       ...
#//|     end
#//|   endtask
#//| endclass
#//

#`define uvm_declare_p_sequencer(SEQUENCER) \
#  SEQUENCER p_sequencer;\
#  virtual function void m_set_p_sequencer();\
#    super.m_set_p_sequencer(); \
#    if( !$cast(p_sequencer, m_sequencer)) \
#        `uvm_fatal("DCLPSQ", \
#        $sformatf("%m %s Error casting p_sequencer, please verify that this sequence/sequence item is intended to execute on this type of sequencer", get_full_name())) \
#  endfunction



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


from uvm.seq.uvm_sequencer import UVMSequencer
from uvm.macros import uvm_component_utils


class apb_sequencer(UVMSequencer):  # (apb_rw)

    def __init__(self, name, parent=None):
        super().__init__(name, parent)


uvm_component_utils(apb_sequencer)




import unittest

from uvm.base.uvm_component import UVMComponent
from uvm.tlm1.uvm_ports import (
    UVMBlockingPutPort, UVMNonBlockingPutPort,
    UVMBlockingGetPort, UVMNonBlockingGetPort,
)

from uvm.tlm1.uvm_imps import (
     UVMNonBlockingPutImp, UVMNonBlockingGetImp
)


class TestUVMPorts(unittest.TestCase):

    def test_non_blocking_put_port(self):
        source = UVMComponent('uvm_ports_comp', None)
        sink = UVMComponent('uvm_ports_comp2', None)

        def try_put(item):
            pass
        setattr(sink, 'try_put', try_put)

        put_port = UVMNonBlockingPutPort('my_port', source)
        put_imp = UVMNonBlockingPutImp('put_imp', sink)
        put_port.connect(put_imp)
        put_port.resolve_bindings()
        put_port.try_put(0x1234)
        # TODO add assert

    def test_blocking_put_port(self):
        comp = UVMComponent('comp', None)
        put_port = UVMBlockingPutPort('my_port', comp)
        self.assertEqual(put_port.is_port(), True)

if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.base import (UVMComponent, UVM_LOW, UVM_HIGH)
from uvm.base.uvm_cmdline_processor import UVMCmdlineProcessor

#------------------------------------------------------------------------------
# UNIT TESTS
#------------------------------------------------------------------------------


class TestUVMComponent(unittest.TestCase):

    def test_name(self):
        from uvm.base.uvm_root import UVMRoot
        uvm_top = UVMRoot.get()
        comp = UVMComponent("my_component", uvm_top)
        self.assertEqual(comp.get_full_name(), "my_component")
        comp_child = UVMComponent("my_child", comp)
        self.assertEqual(comp_child.get_full_name(), "my_component.my_child")

    def test_children(self):
        comp = UVMComponent("parent", None)
        child = UVMComponent("child", comp)
        self.assertEqual(child.get_parent(), comp)
        self.assertEqual(comp.get_first_child(), child)
        self.assertEqual(comp.get_num_children(), 1)
        self.assertEqual(comp.get_child("child"), child)
        self.assertEqual(comp.has_child("child"), True)
        self.assertEqual(comp.has_child("xchild"), False)
        children = []
        comp.get_children(children)
        self.assertEqual(children[0], child)

    def test_lookup(self):
        comp = UVMComponent("parent1", None)
        child = UVMComponent("child", comp)
        sub_child = UVMComponent("sub_child", child)

        c1 = comp.lookup('child')
        sc = comp.lookup('child.sub_child')
        self.assertEqual(c1, child)
        self.assertEqual(sub_child, sc)

    def test_errors(self):
        comp = UVMComponent("parent12345", None)
        child = comp.get_first_child()
        self.assertIsNone(child)
        child = comp.get_next_child()
        self.assertIsNone(child)

    #def test_clp_args(self):
    #    # "+uvm_set_verbosity=<comp>,<id>,<verbosity>,<phase|time>,<offset>"
    #    UVMCmdlineProcessor.m_test_mode = True
    #    UVMCmdlineProcessor.m_test_plusargs = {
    #        "uvm_set_verbosity": "*child1,_ALL_,UVM_LOW,,0",
    #        "uvm_set_verbosity": "parent1,_ALL_,UVM_HIGH,,0",
    #        "ABC": "123"
    #    }
    #    comp = UVMComponent("parent2", None)
    #    child = UVMComponent("child1", comp)
    #    child2 = UVMComponent("child2", comp)
    #    self.assertEqual(child.get_report_verbosity_level(), UVM_LOW)
    #    self.assertEqual(comp.get_report_verbosity_level(), UVM_HIGH)

    #    UVMCmdlineProcessor.m_test_mode = False


if __name__ == '__main__':
    unittest.main()



#// 
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import *
from apb.apb_rw import *
from vip import *


class apb2txrx(UVMComponent):
    #
    #   uvm_analysis_imp#(apb_rw, apb2txrx) apb
    #
    #   uvm_analysis_port#(vip_tr) tx_ap
    #   uvm_analysis_port#(vip_tr) rx_ap
    #

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.apb   = UVMAnalysisImp("apb", self)
        self.tx_ap = UVMAnalysisPort("tx_ap", self)
        self.rx_ap = UVMAnalysisPort("rx_ap", self)
        self.radd =  0x0100
        self.wadd =  0x0100


    def build_phase(self, phase):
        arr = []
        if UVMConfigDb.get(self, "", "radd", arr):
            self.radd = arr[0]
        arr = []
        if UVMConfigDb.get(self, "", "wadd", arr):
            self.wadd = arr[0]


    def write(self, rw):
        # R from 'radd' are Rx characters
        if (rw.kind == apb_rw.READ and rw.addr == self.radd):
            tr = vip_tr.type_id.create("rx",None,self.get_full_name())
            tr.chr = rw.data
            self.rx_ap.write(tr)
        
        # W from 'wadd' are Tx characters
        if (rw.kind == apb_rw.WRITE  and  rw.addr == self.wadd):
            tr = vip_tr.type_id.create("tx",None,self.get_full_name())
            tr.chr = rw.data
            self.tx_ap.write(tr)


uvm_component_utils(apb2txrx)



#//
#//----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   Copyright 2019      Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------------
#

from .uvm_set_get_dap_base import uvm_set_get_dap_base
from ..macros.uvm_object_defines import uvm_object_utils
from ..macros.uvm_message_defines import *
from ..base.sv import sv
from ..base.uvm_globals import uvm_check_output_args

ERR_MSG1 = "Attempt to get value on '%s', but the data access policy forbits calling 'get' prior to calling 'set' or 'try_set'!"

#// Class: uvm_set_before_get_dap
#// Provides a 'Set Before Get' Data Access Policy.
#//
#// The 'Set Before Get' Data Access Policy enforces that the value must
#// be written at ~least~ once before it is read.  This DAP can be used to
#// pass shared information to multiple components during standard configuration,
#// even if that information hasn't yet been determined.
#//
#// Such DAP objects can be useful for passing a 'placeholder' reference, before
#// the information is actually available.  A good example of this would be
#// the virtual sequencer:
#//
#//| typedef uvm_set_before_get_dap#(uvm_sequencer_base) seqr_dap_t;
#//| virtual_seqeuncer_type virtual_sequencer;
#//| agent_type my_agent;
#//| seqr_dap_t seqr_dap;
#//|
#//| function void my_env::build_phase(uvm_phase phase);
#//|   seqr_dap = seqr_dap_t::type_id::create("seqr_dap");
#//|   // Pass the DAP, because we don't have a reference to the
#//|   // real sequencer yet...
#//|   uvm_config_db#(seqr_dap_t)::set(this, "virtual_sequencer", "seqr_dap", seqr_dap);
#//|
#//|   // Create the virtual sequencer
#//|   virtual_sequencer = virtual_sequencer_type::type_id::create("virtual_sequencer", this);
#//|
#//|   // Create the agent
#//|   agent = agent_type::type_id::create("agent", this);
#//| endfunction
#//|
#//| function void my_env::connect_phase(uvm_phase phase);
#//|   // Now that we know the value is good, we can set it
#//|   seqr_dap.set(agent.sequencer);
#//| endfunction
#//
#// In the example above, the environment didn't have a reference to the
#// agent's sequencer yet, because the agent hadn't executed its ~build_phase~.
#// The environment needed to give the virtual sequencer a "Set before get" DAP
#// so that the virtual sequencer (and any sequences one it), could ~eventually~
#// see the agent's sequencer, when the reference was finally available.  If
#// the virtual sequencer (or any sequences on it) attempted to 'get' the
#// reference to the agent's sequencer ~prior~ to the environment assigning it,
#// an error would have been reported.
#
#class uvm_set_before_get_dap#(type T=int) extends uvm_set_get_dap_base#(T);
class uvm_set_before_get_dap(uvm_set_get_dap_base):
    #
    #   // Used for self-references
    #   typedef uvm_set_before_get_dap#(T) this_type;
    #
    #   // Function: new
    #   // Constructor
    def __init__(self, name="unnamed-uvm_set_before_get_dap#(T)"):
        uvm_set_get_dap_base.__init__(self, name)
        self.m_set = False
        self.m_value = None

    #   // Group: Set/Get Interface
    #
    #   // Function: set
    #   // Updates the value stored within the DAP.
    #   //
    def set(self, value):
        self.m_set = True
        self.m_value = value

    #   // Function: try_set
    #   // Attempts to update the value stored within the DAP.
    #   //
    #   // ~try_set~ will always return a 1.
    #   virtual function bit try_set(T value);
    #      set(value);
    #      return 1;
    #   endfunction : try_set
    #

    #   // Function: get
    #   // Returns the current value stored within the DAP.
    #   //
    #   // If 'get' is called before a call to <set> or <try_set>, then
    #   // an error will be reported.
    def get(self):
        if self.m_set is False:
            uvm_error("UVM/SET_BEFORE_GET_DAP/NO_SET", sv.sformatf(ERR_MSG1, self.get_full_name()))
        return self.m_value

    #
    #   // Function: try_get
    #   // Attempts to retrieve the current value stored within the DAP
    #   //
    #   // If the value has not been 'set', then try_get will return a 0,
    #   // otherwise it will return a 1, and set ~value~ to the current
    #   // value stored within the DAP.
    def try_get(self, value):
        uvm_check_output_args([value])
        if self.m_set is False:
            return 0
        else:
            value.append(self.m_value)
            return 1

    #
    #   // Group: Introspection
    #   //
    #   // The ~uvm_set_before_get_dap~ cannot support the standard UVM
    #   // instrumentation methods (~copy~, ~clone~, ~pack~ and
    #   // ~unpack~), due to the fact that they would potentially
    #   // violate the access policy.
    #   //
    #   // A call to any of these methods will result in an error.
    #
    #   virtual function void do_copy(uvm_object rhs);
    #      `uvm_error("UVM/SET_BEFORE_GET_DAP/CPY",
    #                 "'copy()' is not supported for 'uvm_set_before_get_dap#(T)'")
    #   endfunction : do_copy
    #
    #   virtual function void do_pack(uvm_packer packer);
    #      `uvm_error("UVM/SET_BEFORE_GET_DAP/PCK",
    #                 "'pack()' is not supported for 'uvm_set_before_get_dap#(T)'")
    #   endfunction : do_pack
    #
    #   virtual function void do_unpack(uvm_packer packer);
    #      `uvm_error("UVM/SET_BEFORE_GET_DAP/UPK",
    #                 "'unpack()' is not supported for 'uvm_set_before_get_dap#(T)'")
    #   endfunction : do_unpack
    #
    #   // Group- Reporting
    #
    #   // Function- convert2string
    #   virtual function string convert2string();
    #      if (m_set)
    #        return $sformatf("(%s) %0p [SET]", `uvm_typename(m_value), m_value);
    #      else
    #        return $sformatf("(%s) %0p [UNSET]", `uvm_typename(m_value), m_value);
    #   endfunction : convert2string
    #
    #   // Function- do_print
    #   virtual function void do_print(uvm_printer printer);
    #      super.do_print(printer);
    #      printer.print_field_int("set_state", m_set, $bits(m_set));
    #      printer.print_generic("value",
    #                            `uvm_typename(m_value),
    #                            0,
    #                            $sformatf("%0p", m_value));
    #
    #   endfunction : do_print
    #
    #endclass // uvm_set_before_get_dap


uvm_object_utils(uvm_set_before_get_dap)



#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------
#
#`define UVM_TASK_ERROR "TLM interface task not implemented"
#`define UVM_FUNCTION_ERROR "TLM interface function not implemented"
#
#//-----------------------------------------------------------------------------
#//
#// CLASS: uvm_tlm_if_base #(T1,T2)
#//
#// This class declares all of the methods of the TLM API.
#//
#// Various subsets of these methods are combined to form primitive TLM
#// interfaces, which are then paired in various ways to form more abstract
#// "combination" TLM interfaces. Components that require a particular interface
#// use ports to convey that requirement. Components that provide a particular
#// interface use exports to convey its availability.
#//
#// Communication between components is established by connecting ports to
#// compatible exports, much like connecting module signal-level output ports to
#// compatible input ports. The difference is that UVM ports and exports bind
#// interfaces (groups of methods), not signals and wires. The methods of the
#// interfaces so bound pass data as whole transactions (e.g. objects).
#// The set of primitive and combination TLM interfaces afford many choices for
#// designing components that communicate at the transaction level.
#// 
#//-----------------------------------------------------------------------------
#
#virtual class uvm_tlm_if_base #(type T1=int, type T2=int);
#
#  // Group: Blocking put
#
#  // Task: put
#  //
#  // Sends a user-defined transaction of type T. 
#  //
#  // Components implementing the put method will block the calling thread if
#  // it cannot immediately accept delivery of the transaction.
#
#  virtual task put( input T1 t );
#    uvm_report_error("put", `UVM_TASK_ERROR, UVM_NONE);
#  endtask
#
#  // Group: Blocking get
#
#  // Task: get
#  //
#  // Provides a new transaction of type T. 
#  //
#  // The calling thread is blocked if the requested transaction cannot be
#  // provided immediately. The new transaction is returned in the provided
#  // output argument. 
#  //
#  // The implementation of get must regard the transaction as consumed.
#  // Subsequent calls to get must return a different transaction instance.
#
#  virtual task get( output T2 t );
#    uvm_report_error("get", `UVM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Group: Blocking peek
#
#  // Task: peek
#  //
#  // Obtain a new transaction without consuming it. 
#  //
#  // If a transaction is available, then it is written to the provided output
#  // argument. If a transaction is not available, then the calling thread is
#  // blocked until one is available. 
#  //
#  // The returned transaction is not consumed. A subsequent peek or get will
#  // return the same transaction.
#
#  virtual task peek( output T2 t );
#    uvm_report_error("peek", `UVM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Group: Non-blocking put
#
#  // Function: try_put
#  //
#  // Sends a transaction of type T, if possible. 
#  //
#  // If the component is ready to accept the transaction argument, then it does
#  // so and returns 1, otherwise it returns 0.
#
#  virtual function bit try_put( input T1 t );
#    uvm_report_error("try_put", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Function: can_put
#  //
#  // Returns 1 if the component is ready to accept the transaction; 0 otherwise.
#
#  virtual function bit can_put();
#    uvm_report_error("can_put", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Group: Non-blocking get
#
#  // Function: try_get
#  //
#  // Provides a new transaction of type T.
#  //
#  // If a transaction is immediately available, then it is written to the output
#  // argument and 1 is returned. Otherwise, the output argument is not modified
#  // and 0 is returned.
#
#  virtual function bit try_get( output T2 t );
#    uvm_report_error("try_get", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Function: can_get
#  //
#  // Returns 1 if a new transaction can be provided immediately upon request,
#  // 0 otherwise.
#
#  virtual function bit can_get();
#    uvm_report_error("can_get", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Group: Non-blocking peek
#
#  // Function: try_peek
#  //
#  // Provides a new transaction without consuming it. 
#  //
#  // If available, a transaction is written to the output argument and 1 is
#  // returned. A subsequent peek or get will return the same transaction. If a
#  // transaction is not available, then the argument is unmodified and 0 is
#  // returned.
#
#  virtual function bit try_peek( output T2 t );
#    uvm_report_error("try_peek", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Function: can_peek
#  //
#  // Returns 1 if a new transaction is available; 0 otherwise.
#
#  virtual function bit can_peek();
#    uvm_report_error("can_ppeek", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Group: Blocking transport
#
#  // Task: transport
#  //
#  // Executes the given request and returns the response in the given output
#  // argument. The calling thread may block until the operation is complete.
#
#  virtual task transport( input T1 req , output T2 rsp );
#    uvm_report_error("transport", `UVM_TASK_ERROR, UVM_NONE);
#  endtask
#
#
#  // Group: Non-blocking transport
#
#  // Task: nb_transport
#  //
#  // Executes the given request and returns the response in the given output
#  // argument. Completion of this operation must occur without blocking.
#  //
#  // If for any reason the operation could not be executed immediately, then
#  // a 0 must be returned; otherwise 1.
#
#  virtual function bit nb_transport(input T1 req, output T2 rsp);
#    uvm_report_error("nb_transport", `UVM_FUNCTION_ERROR, UVM_NONE);
#    return 0;
#  endfunction
#
#
#  // Group: Analysis
#
#  // Function: write
#  //
#  // Broadcasts a user-defined transaction of type T to any number of listeners.
#  // The operation must complete without blocking. 
#
#  virtual function void write( input T1 t );
#    uvm_report_error("write", `UVM_FUNCTION_ERROR, UVM_NONE);
#  endfunction
#
#endclass
#



#----------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010-2011 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
# ----------------------------------------------------------------------

import cocotb
from cocotb.utils import get_sim_time
from cocotb.triggers import Timer
from cocotb.clock import Clock

from uvm.macros import uvm_component_utils
from uvm.base.uvm_globals import run_test
from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_object_globals import *
from uvm.base.uvm_debug import uvm_debug, UVMDebug

MARKER = "===============>>>>>>>>>"


class MyComponent(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
    
    
    async def run_phase(self, phase):
        self.uvm_report_info("component", "before raising objection", UVM_MEDIUM)
        phase.raise_objection(self)
        UVMDebug.DEBUG = False
        self.uvm_report_info("component", "after raising objection", UVM_MEDIUM)
        await Timer(1, 'ns')
        uvm_debug(self, 'run_phase', "Objection raised run_phase " +
                self.get_name())
        await Timer(2, 'ns')
        self.uvm_report_info("component", "hello out there!", UVM_MEDIUM)
        await Timer(1000, 'ns')
        self.uvm_report_info("component", "finishing up!", UVM_MEDIUM)
        phase.drop_objection(self)

uvm_component_utils(MyComponent)


@cocotb.test()
async def simple_trivial_component_test(dut):
    cc = MyComponent("Top", None)
    print(cc.get_full_name())

    # tpoikela: Required by ghdl
    cocotb.fork(Clock(dut.clk, 1, "NS").start())
    await run_test()
    sim_time = get_sim_time()
    await Timer(501, 'ns')

    if sim_time == 0:
        raise Exception('Sim time has not progressed')
    else:
        sim_time = get_sim_time('ns')
        print("Sim time at the end is " + str(get_sim_time('ns')))
        if str(sim_time) != "1504.0":
            msg = "Exp: 1504.0, got: " + str(sim_time)
            raise Exception('Wrong sim time at the end: ' + msg)




class UVMDefaultServer:

    def __init__(self):
        self.m_quit_count = 0
        self.m_max_quit_count = 0
        self.max_quit_overridable = True
        self.m_severity_count = {}
        self.m_id_count = {}
        ## uvm_tr_database m_message_db;
        ## uvm_tr_stream m_streams[string][string]; // ro.name,rh.name




#//----------------------------------------------------------------------
#// Copyright 2010-2012 Mentor Graphics Corporation
#// Copyright 2014 Semifore
#// Copyright 2010-2012 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2013-2018 NVIDIA Corporation
#// Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
TLM Generic Payload & Extensions

 The Generic Payload transaction represents a generic
 bus read/write access. It is used as the default transaction in
 TLM2 blocking and nonblocking transport interfaces.
"""

from ..seq.uvm_sequence_item import *
from ..macros import *
from ..base.uvm_object import UVMObject


from enum import Enum, auto, IntEnum
from uvm.seq.uvm_sequence_item import UVMSequenceItem
from uvm.base.uvm_object import UVMObject


class uvm_tlm_command_e(Enum):
    """
    Enum uvm_tlm_command_e
    Command attribute type definition

    UVM_TLM_READ_COMMAND      - Bus read operation

    UVM_TLM_WRITE_COMMAND     - Bus write operation

    UVM_TLM_IGNORE_COMMAND    - No bus operation.
    """
    UVM_TLM_READ_COMMAND = auto()
    UVM_TLM_WRITE_COMMAND = auto()
    UVM_TLM_IGNORE_COMMAND = auto()

UVM_TLM_READ_COMMAND = uvm_tlm_command_e.UVM_TLM_READ_COMMAND
UVM_TLM_WRITE_COMMAND = uvm_tlm_command_e.UVM_TLM_WRITE_COMMAND
UVM_TLM_IGNORE_COMMAND = uvm_tlm_command_e.UVM_TLM_IGNORE_COMMAND


class uvm_tlm_response_status_e(IntEnum):
    """
    Enum uvm_tlm_response_status_e
    Response status attribute type definition

    UVM_TLM_OK_RESPONSE                - Bus operation completed successfully

    UVM_TLM_INCOMPLETE_RESPONSE        - Transaction was not delivered to target

    UVM_TLM_GENERIC_ERROR_RESPONSE     - Bus operation had an error

    UVM_TLM_ADDRESS_ERROR_RESPONSE     - Invalid address specified

    UVM_TLM_COMMAND_ERROR_RESPONSE     - Invalid command specified

    UVM_TLM_BURST_ERROR_RESPONSE       - Invalid burst specified

    UVM_TLM_BYTE_ENABLE_ERROR_RESPONSE - Invalid byte enabling specified
    """

    OK_RESPONSE = 1
    INCOMPLETE_RESPONSE = 0
    GENERIC_ERROR_RESPONSE = -1
    ADDRESS_ERROR_RESPONSE = -2
    COMMAND_ERROR_RESPONSE = -3
    BURST_ERROR_RESPONSE = -4
    BYTE_ENABLE_ERROR_RESPONSE = -5


UVM_TLM_OK_RESPONSE=uvm_tlm_response_status_e.OK_RESPONSE
UVM_TLM_INCOMPLETE_RESPONSE=uvm_tlm_response_status_e.INCOMPLETE_RESPONSE
UVM_TLM_GENERIC_ERROR_RESPONSE=uvm_tlm_response_status_e.GENERIC_ERROR_RESPONSE
UVM_TLM_ADDRESS_ERROR_RESPONSE=uvm_tlm_response_status_e.ADDRESS_ERROR_RESPONSE
UVM_TLM_COMMAND_ERROR_RESPONSE=uvm_tlm_response_status_e.COMMAND_ERROR_RESPONSE
UVM_TLM_BURST_ERROR_RESPONSE=uvm_tlm_response_status_e.BURST_ERROR_RESPONSE
UVM_TLM_BYTE_ENABLE_ERROR_RESPONSE=uvm_tlm_response_status_e.BYTE_ENABLE_ERROR_RESPONSE

#//-----------------------
#// Group -- NODOCS -- Generic Payload
#//-----------------------

#// @uvm-ieee 1800.2-2017 auto 12.3.4.2.1
class UVMTLMGenericPayload(UVMSequenceItem):
    """
    This class provides a transaction definition commonly used in
    memory-mapped bus-based systems.  It's intended to be a general
    purpose transaction class that lends itself to many applications. The
    class is derived from uvm_sequence_item which enables it to be
    generated in sequences and transported to drivers through sequencers.

    :ivar int m_address: Address for the bus operation.
      Should be set or read using the <set_address> and <get_address>
      methods. The variable should be used only when constraining.

      For a read command or a write command, the target shall
      interpret the current value of the address attribute as the start
      address in the system memory map of the contiguous block of data
      being read or written.
      The address associated with any given byte in the data array is
      dependent upon the address attribute, the array index, the
      streaming width attribute, the endianness and the width of the physical bus.

      If the target is unable to execute the transaction with
      the given address attribute (because the address is out-of-range,
      for example) it shall generate a standard error response. The
      recommended response status is ~UVM_TLM_ADDRESS_ERROR_RESPONSE~.
    """

    #   rand bit [63:0]             m_address


    #   // Variable -- NODOCS -- m_command
    #   //
    #   // Bus operation type.
    #   // Should be set using the <set_command>, <set_read> or <set_write> methods
    #   // and read using the <get_command>, <is_read> or <is_write> methods.
    #   // The variable should be used only when constraining.
    #   //
    #   // If the target is unable to execute a read or write command, it
    #   // shall generate a standard error response. The
    #   // recommended response status is UVM_TLM_COMMAND_ERROR_RESPONSE.
    #   //
    #   // On receipt of a generic payload transaction with the command
    #   // attribute equal to UVM_TLM_IGNORE_COMMAND, the target shall not execute
    #   // a write command or a read command not modify any data.
    #   // The target may, however, use the value of any attribute in
    #   // the generic payload, including any extensions.
    #   //
    #   // The command attribute shall be set by the initiator, and shall
    #   // not be overwritten by any interconnect
    #   //
    #   rand uvm_tlm_command_e          m_command



    #   // Variable -- NODOCS -- m_data
    #   //
    #   // Data read or to be written.
    #   // Should be set and read using the <set_data> or <get_data> methods
    #   // The variable should be used only when constraining.
    #   //
    #   // For a read command or a write command, the target shall copy data
    #   // to or from the data array, respectively, honoring the semantics of
    #   // the remaining attributes of the generic payload.
    #   //
    #   // For a write command or UVM_TLM_IGNORE_COMMAND, the contents of the
    #   // data array shall be set by the initiator, and shall not be
    #   // overwritten by any interconnect component or target. For a read
    #   // command, the contents of the data array shall be overwritten by the
    #   // target (honoring the semantics of the byte enable) but by no other
    #   // component.
    #   //
    #   // Unlike the OSCI TLM-2.0 LRM, there is no requirement on the endiannes
    #   // of multi-byte data in the generic payload to match the host endianness.
    #   // Unlike C++, it is not possible in SystemVerilog to cast an arbitrary
    #   // data type as an array of bytes. Therefore, matching the host
    #   // endianness is not necessary. In contrast, arbitrary data types may be
    #   // converted to and from a byte array using the streaming operator and
    #   // <uvm_object> objects may be further converted using the
    #   // <uvm_object::pack_bytes()> and <uvm_object::unpack_bytes()> methods.
    #   // All that is required is that a consistent mechanism is used to
    #   // fill the payload data array and later extract data from it.
    #   //
    #   // Should a generic payload be transferred to/from a SystemC model,
    #   // it will be necessary for any multi-byte data in that generic payload
    #   // to use/be interpreted using the host endianness.
    #   // However, this process is currently outside the scope of this standard.
    #   //
    #   rand byte unsigned             m_data[]


    #   // Variable -- NODOCS -- m_length
    #   //
    #   // The number of bytes to be copied to or from the <m_data> array,
    #   // inclusive of any bytes disabled by the <m_byte_enable> attribute.
    #   //
    #   // The data length attribute shall be set by the initiator,
    #   // and shall not be overwritten by any interconnect component or target.
    #   //
    #   // The data length attribute shall not be set to 0.
    #   // In order to transfer zero bytes, the <m_command> attribute
    #   // should be set to <UVM_TLM_IGNORE_COMMAND>.
    #   //
    #   rand int unsigned           m_length
    #
    #
    #   // Variable -- NODOCS -- m_response_status
    #   //
    #   // Status of the bus operation.
    #   // Should be set using the <set_response_status> method
    #   // and read using the <get_response_status>, <get_response_string>,
    #   // <is_response_ok> or <is_response_error> methods.
    #   // The variable should be used only when constraining.
    #   //
    #   // The response status attribute shall be set to
    #   // UVM_TLM_INCOMPLETE_RESPONSE by the initiator, and may
    #   // be overwritten by the target. The response status attribute
    #   // should not be overwritten by any interconnect
    #   // component, because the default value UVM_TLM_INCOMPLETE_RESPONSE
    #   // indicates that the transaction was not delivered to the target.
    #   //
    #   // The target may set the response status attribute to UVM_TLM_OK_RESPONSE
    #   // to indicate that it was able to execute the command
    #   // successfully, or to one of the five error responses
    #   // to indicate an error. The target should choose the appropriate
    #   // error response depending on the cause of the error.
    #   // If a target detects an error but is unable to select a specific
    #   // error response, it may set the response status to
    #   // UVM_TLM_GENERIC_ERROR_RESPONSE.
    #   //
    #   // The target shall be responsible for setting the response status
    #   // attribute at the appropriate point in the
    #   // lifetime of the transaction. In the case of the blocking
    #   // transport interface, this means before returning
    #   // control from b_transport. In the case of the non-blocking
    #   // transport interface and the base protocol, this
    #   // means before sending the BEGIN_RESP phase or returning a value of UVM_TLM_COMPLETED.
    #   //
    #   // It is recommended that the initiator should always check the
    #   // response status attribute on receiving a
    #   // transition to the BEGIN_RESP phase or after the completion of
    #   // the transaction. An initiator may choose
    #   // to ignore the response status if it is known in advance that the
    #   // value will be UVM_TLM_OK_RESPONSE,
    #   // perhaps because it is known in advance that the initiator is
    #   // only connected to targets that always return
    #   // UVM_TLM_OK_RESPONSE, but in general this will not be the case. In
    #   // other words, the initiator ignores the
    #   // response status at its own risk.
    #   //
    #   rand uvm_tlm_response_status_e  m_response_status
    #
    #
    #   // Variable -- NODOCS -- m_dmi
    #   //
    #   // DMI mode is not yet supported in the UVM TLM2 subset.
    #   // This variable is provided for completeness and interoperability
    #   // with SystemC.
    #   //
    #   bit m_dmi
    #
    #
    #   // Variable -- NODOCS -- m_byte_enable
    #   //
    #   // Indicates valid <m_data> array elements.
    #   // Should be set and read using the <set_byte_enable> or <get_byte_enable> methods
    #   // The variable should be used only when constraining.
    #   //
    #   // The elements in the byte enable array shall be interpreted as
    #   // follows. A value of 8'h00 shall indicate that that
    #   // corresponding byte is disabled, and a value of 8'hFF shall
    #   // indicate that the corresponding byte is enabled.
    #   //
    #   // Byte enables may be used to create burst transfers where the
    #   // address increment between each beat is
    #   // greater than the number of significant bytes transferred on each
    #   // beat, or to place words in selected byte
    #   // lanes of a bus. At a more abstract level, byte enables may be
    #   // used to create "lacy bursts" where the data array of the generic
    #   // payload has an arbitrary pattern of holes punched in it.
    #   //
    #   // The byte enable mask may be defined by a small pattern applied
    #   // repeatedly or by a large pattern covering the whole data array.
    #   // The byte enable array may be empty, in which case byte enables
    #   // shall not be used for the current transaction.
    #   //
    #   // The byte enable array shall be set by the initiator and shall
    #   // not be overwritten by any interconnect component or target.
    #   //
    #   // If the byte enable pointer is not empty, the target shall either
    #   // implement the semantics of the byte enable as defined below or
    #   // shall generate a standard error response. The recommended response
    #   // status is UVM_TLM_BYTE_ENABLE_ERROR_RESPONSE.
    #   //
    #   // In the case of a write command, any interconnect component or
    #   // target should ignore the values of any disabled bytes in the
    #   // <m_data> array. In the case of a read command, any interconnect
    #   // component or target should not modify the values of disabled
    #   // bytes in the <m_data> array.
    #   //
    #   rand byte unsigned          m_byte_enable[]
    #
    #
    #   // Variable -- NODOCS -- m_byte_enable_length
    #   //
    #   // The number of elements in the <m_byte_enable> array.
    #   //
    #   // It shall be set by the initiator, and shall not be overwritten
    #   // by any interconnect component or target.
    #   //
    #   rand int unsigned m_byte_enable_length
    #
    #
    #   // Variable -- NODOCS -- m_streaming_width
    #   //
    #   // Number of bytes transferred on each beat.
    #   // Should be set and read using the <set_streaming_width> or
    #   // <get_streaming_width> methods
    #   // The variable should be used only when constraining.
    #   //
    #   // Streaming affects the way a component should interpret the data
    #   // array. A stream consists of a sequence of data transfers occurring
    #   // on successive notional beats, each beat having the same start
    #   // address as given by the generic payload address attribute. The
    #   // streaming width attribute shall determine the width of the stream,
    #   // that is, the number of bytes transferred on each beat. In other
    #   // words, streaming affects the local address associated with each
    #   // byte in the data array. In all other respects, the organization of
    #   // the data array is unaffected by streaming.
    #   //
    #   // The bytes within the data array have a corresponding sequence of
    #   // local addresses within the component accessing the generic payload
    #   // transaction. The lowest address is given by the value of the
    #   // address attribute. The highest address is given by the formula
    #   // address_attribute + streaming_width - 1. The address to or from
    #   // which each byte is being copied in the target shall be set to the
    #   // value of the address attribute at the start of each beat.
    #   //
    #   // With respect to the interpretation of the data array, a single
    #   // transaction with a streaming width shall be functionally equivalent
    #   // to a sequence of transactions each having the same address as the
    #   // original transaction, each having a data length attribute equal to
    #   // the streaming width of the original, and each with a data array
    #   // that is a different subset of the original data array on each
    #   // beat. This subset effectively steps down the original data array
    #   // maintaining the sequence of bytes.
    #   //
    #   // A streaming width of 0 indicates that a streaming transfer
    #   // is not required. it is equivalent to a streaming width
    #   // value greater than or equal to the size of the <m_data> array.
    #   //
    #   // Streaming may be used in conjunction with byte enables, in which
    #   // case the streaming width would typically be equal to the byte
    #   // enable length. It would also make sense to have the streaming width
    #   // a multiple of the byte enable length. Having the byte enable length
    #   // a multiple of the streaming width would imply that different bytes
    #   // were enabled on each beat.
    #   //
    #   // If the target is unable to execute the transaction with the
    #   // given streaming width, it shall generate a standard error
    #   // response. The recommended response status is
    #   // TLM_BURST_ERROR_RESPONSE.
    #   //
    #   rand int unsigned m_streaming_width


    #   protected uvm_tlm_extension_base m_extensions [uvm_tlm_extension_base]
    #   local rand uvm_tlm_extension_base m_rand_exts[]





    #  // Function -- NODOCS -- new
    #  //
    #  // Create a new instance of the generic payload.  Initialize all the
    #  // members to their default values.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.3
    def __init__(self, name=""):
        super().__init__(name)
        self.m_address = 0
        self.m_command = UVM_TLM_IGNORE_COMMAND
        self.m_length = 0
        self.m_response_status = UVM_TLM_INCOMPLETE_RESPONSE
        self.m_dmi = 0
        self.m_byte_enable_length = 0
        self.m_streaming_width = 0
        self.m_data = []
        #self.rand('m_data')


    #  // Function- do_print
    #  //
    #  def void do_print(self,uvm_printer printer):
    #    byte unsigned be
    #    super().do_print(printer)
    #    printer.print_field_int     ("address", m_address, 64, UVM_HEX)
    #    printer.print_generic ("command", "uvm_tlm_command_e", 32, m_command.name())
    #    printer.print_generic ("response_status", "uvm_tlm_response_status_e",
    #                             32, m_response_status.name())
    #    printer.print_field_int     ("streaming_width", m_streaming_width, 32, UVM_HEX)
    #
    #    printer.print_array_header("data", m_length, "darray(byte)")
    #    for (int i=0; i < m_length  and  i < m_data.size(); i++):
    #      if (m_byte_enable_length):
    #        be = m_byte_enable[i % m_byte_enable_length]
    #        printer.print_generic (sv.sformatf("[%0d]",i), "byte", 8,
    #            sv.sformatf("'h%h%s",m_data[i],((be== 0xFF) ? "" : " x")))
    #      end
    #      else
    #        printer.print_generic (sv.sformatf("[%0d]",i), "byte", 8,
    #                               sv.sformatf("'h%h",m_data[i]))
    #    end
    #    printer.print_array_footer()
    #
    #    begin
    #    string name
    #    printer.print_array_header("extensions", m_extensions.num(), "aa(obj,obj)")
    #    foreach (m_extensions[ext_]):
    #      uvm_tlm_extension_base ext = m_extensions[ext_]
    #      name = {"[",ext.get_name(),"]"}
    #      printer.print_object(name, ext, "[")
    #    end
    #    printer.print_array_footer()
    #    end
    #  endfunction



    #  // Function- do_copy
    #  //
    #  def void do_copy(self,uvm_object rhs):
    #    uvm_tlm_generic_payload gp
    #    super().do_copy(rhs)
    #    sv.cast(gp, rhs)
    #    m_address            = gp.m_address
    #    m_command            = gp.m_command
    #    m_data               = gp.m_data
    #    m_dmi                = gp.m_dmi
    #    m_length             = gp.m_length
    #    m_response_status    = gp.m_response_status
    #    m_byte_enable        = gp.m_byte_enable
    #    m_streaming_width    = gp.m_streaming_width
    #    m_byte_enable_length = gp.m_byte_enable_length
    #
    #    m_extensions.delete()
    #    foreach (gp.m_extensions[ext])
    #       sv.cast(m_extensions[ext], gp.m_extensions[ext].clone())
    #
    #  endfunction


    #`define m_uvm_tlm_fast_compare_int(VALUE,RADIX,NAME="") \
    #  if ( (!comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()))  and  \
    #      ((VALUE) != (gp.VALUE)) ): \
    #     string name = (NAME == "") ? `"VALUE`" : NAME; \
    #        comparer.compare_field_int(name , VALUE, gp.VALUE, sv.bits(VALUE), RADIX)  # cast to 'void' removed \
    #  end


    #`define m_uvm_tlm_fast_compare_enum(VALUE,TYPE,NAME="") \
    #  if ( (!comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()))  and  \
    #      ((VALUE) != (gp.VALUE)) ): \
    #        string name = (NAME == "") ? `"VALUE`" : NAME; \
    #        void'( comparer.compare_string(name, \
    #				sv.sformatf("%s'(%s)", `"TYPE`", VALUE.name()), \
    #				sv.sformatf("%%s)", `"TYPE`", gp.VALUE.name()))   # cast to 's' removed \
    #  end

    #  // Function: do_compare
    #  // Compares this generic payload to ~rhs~.
    #  //
    #  // The <do_compare> method compares the fields of this instance to
    #  // to those of ~rhs~.  All fields are compared, however if byte
    #  // enables are being used, then non-enabled bytes of data are
    #  // skipped.
    #  def bit do_compare(self,uvm_object rhs, uvm_comparer comparer):
    #    uvm_tlm_generic_payload gp
    #    do_compare = super().do_compare(rhs, comparer)
    #    sv.cast(gp, rhs)
    #
    #    `m_uvm_tlm_fast_compare_int(m_address, UVM_HEX)
    #    `m_uvm_tlm_fast_compare_enum(m_command, uvm_tlm_command_e)
    #    `m_uvm_tlm_fast_compare_int(m_length, UVM_UNSIGNED)
    #    `m_uvm_tlm_fast_compare_int(m_dmi, UVM_BIN)
    #    `m_uvm_tlm_fast_compare_int(m_byte_enable_length, UVM_UNSIGNED)
    #    `m_uvm_tlm_fast_compare_enum(m_response_status, uvm_tlm_response_status_e)
    #    `m_uvm_tlm_fast_compare_int(m_streaming_width, UVM_UNSIGNED)
    #
    #    if ( (!comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()))  and
    #	 m_byte_enable_length == gp.m_byte_enable_length ):
    #
    #       for (int i=0; i < m_byte_enable_length  and  i < m_byte_enable.size(); i++):
    #	  `m_uvm_tlm_fast_compare_int(m_byte_enable[i], UVM_HEX, sv.sformatf("m_byte_enable[%0d]", i))
    #       end
    #    end
    #
    #    if ( (!comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()))  and
    #	 m_length == gp.m_length ):
    #
    #        byte unsigned be
    #        for (int i=0; i < m_length  and  i < m_data.size(); i++):
    #          if (m_byte_enable_length):
    #            be = m_byte_enable[i % m_byte_enable_length]
    #          end
    #          else begin
    #	    be = 8'hFF
    #          end
    #	   `m_uvm_tlm_fast_compare_int(m_data[i] & be, UVM_HEX, sv.sformatf("m_data[%0d] & %0x", i, be))
    #        end
    #     end
    #
    #     if ( !comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()) )
    #      foreach (m_extensions[ext_]):
    #         uvm_tlm_extension_base ext = ext_
    #         uvm_tlm_extension_base rhs_ext = gp.m_extensions.exists(ext) ?
    #                                          gp.m_extensions[ext] : None
    #
    #	   void'(comparer.compare_object(ext.get_name(),
    #                                         m_extensions[ext], rhs_ext))
    #
    #	 if ( !comparer.get_threshold()  or  (comparer.get_result() < comparer.get_threshold()) )
    #	   break
    #
    #      end
    #
    #    if (comparer.get_result()):
    #       string msg = sv.sformatf("GP miscompare between '%s' and '%s':\nlhs = %s\nrhs = %s",
    #			      get_full_name(), gp.get_full_name(),
    #			      self.convert2string(), gp.convert2string())
    #       comparer.print_msg(msg)
    #    end
    #
    #    return (comparer.get_result() == 0)
    #  endfunction
    #
    #`undef m_uvm_tlm_fast_compare_int
    #`undef m_uvm_tlm_fast_compare_enum


    #  // Function: do_pack
    #  // Packs the fields of the payload in ~packer~.
    #  //
    #  // Fields are packed in the following order:
    #  // - <m_address>
    #  // - <m_command>
    #  // - <m_length>
    #  // - <m_dmi>
    #  // - <m_data> (if <m_length> is greater than 0)
    #  // - <m_response_status>
    #  // - <m_byte_enable_length>
    #  // - <m_byte_enable> (if <m_byte_enable_length> is greater than 0)
    #  // - <m_streaming_width>
    #  //
    #  // Only <m_length> bytes of the <m_data> array are packed,
    #  // and a fatal message is generated if ~m_data.size()~ is less
    #  // than <m_length>.  The same is true for <m_byte_enable_length>
    #  // and <m_byte_enable>.
    #  //
    #  // Note: The extensions are not packed.
    #  def void do_pack(self,uvm_packer packer):
    #    super().do_pack(packer)
    #    if (m_length > m_data.size())
    #       `uvm_fatal("PACK_DATA_ARR",
    #         sv.sformatf("Data array m_length property (%0d) greater than m_data.size (%0d)",
    #         m_length,m_data.size()))
    #    if (m_byte_enable_length > m_byte_enable.size())
    #       `uvm_fatal("PACK_DATA_ARR",
    #         sv.sformatf("Data array m_byte_enable_length property (%0d) greater than m_byte_enable.size (%0d)",
    #         m_byte_enable_length,m_byte_enable.size()))
    #    `uvm_pack_intN  (m_address,64)
    #    `uvm_pack_enumN (m_command,32)
    #    `uvm_pack_intN  (m_length,32)
    #    `uvm_pack_intN  (m_dmi,1)
    #    for (int i=0; i<m_length; i++)
    #      `uvm_pack_intN(m_data[i],8)
    #    `uvm_pack_enumN (m_response_status,32)
    #    `uvm_pack_intN  (m_byte_enable_length,32)
    #    for (int i=0; i<m_byte_enable_length; i++)
    #      `uvm_pack_intN(m_byte_enable[i],8)
    #    `uvm_pack_intN  (m_streaming_width,32)
    #
    #  endfunction


    #  // Function: do_unpack
    #  // Unpacks the fields of the payload from ~packer~.
    #  //
    #  // The <m_data>/<m_byte_enable> arrays are reallocated if the
    #  // new size is greater than their current size; otherwise the
    #  // existing array allocations are kept.
    #  //
    #  // Note: The extensions are not unpacked.
    #  def void do_unpack(self,uvm_packer packer):
    #    super().do_unpack(packer)
    #    `uvm_unpack_intN  (m_address,64)
    #    `uvm_unpack_enumN (m_command, 32, uvm_tlm_command_e)
    #    `uvm_unpack_intN  (m_length,32)
    #    `uvm_unpack_intN (m_dmi, 1)
    #    if (m_data.size() < m_length)
    #      m_data = new[m_length]
    #    foreach (m_data[i])
    #      `uvm_unpack_intN(m_data[i],8)
    #    `uvm_unpack_enumN (m_response_status, 32, uvm_tlm_response_status_e)
    #    `uvm_unpack_intN  (m_byte_enable_length,32)
    #    if (m_byte_enable.size() < m_byte_enable_length)
    #      m_byte_enable = new[m_byte_enable_length]
    #    for (int i=0; i<m_byte_enable_length; i++)
    #      `uvm_unpack_intN(m_byte_enable[i],8)
    #    `uvm_unpack_intN  (m_streaming_width,32)
    #
    #  endfunction


    #  // Function- do_record
    #  //
    #  def void do_record(self,uvm_recorder recorder):
    #    if (!is_recording_enabled())
    #      return
    #    super().do_record(recorder)
    #    `uvm_record_int("address",m_address,sv.bits(m_address))
    #    `uvm_record_string("command",m_command.name())
    #    `uvm_record_int("data_length",m_length,sv.bits(m_length))
    #    `uvm_record_int("byte_enable_length",m_byte_enable_length,sv.bits(m_byte_enable_length))
    #    `uvm_record_string("response_status",m_response_status.name())
    #    `uvm_record_int("streaming_width",m_streaming_width,sv.bits(m_streaming_width))
    #
    #    for (int i=0; i < m_length; i++)
    #      `uvm_record_int(sv.sformatf("\\data[%0d] ", i), m_data[i], sv.bits(m_data[i]))
    #
    #    for (int i=0; i < m_byte_enable_length; i++)
    #      `uvm_record_int(sv.sformatf("\\byte_en[%0d] ", i), m_byte_enable[i], sv.bits(m_byte_enable[i]))
    #
    #    foreach (m_extensions[ext])
    #      recorder.record_object(ext.get_name(),m_extensions[ext])
    #  endfunction


    #// Function- convert2string
    #//
    def convert2string(self):
        msg = "%s %s [0x%16x] = " % (super().convert2string(), str(self.m_command), self.m_address)

        for i in range(self.m_length):
            if (self.m_byte_enable_length == 0 or (self.m_byte_enable[i % self.m_byte_enable_length] == 0xFF)):
                s = " %02x" % (self.m_data[i])
            else:
                s = " --"
            msg += s

        msg += " (status=" + self.get_response_string() + ")"

        return msg;

    #  //--------------------------------------------------------------------
    #  // Group -- NODOCS -- Accessors
    #  //
    #  // The accessor functions let you set and get each of the members of the
    #  // generic payload. All of the accessor methods are virtual. This implies
    #  // a slightly different use model for the generic payload than
    #  // in SystemC. The way the generic payload is defined in SystemC does
    #  // not encourage you to create new transaction types derived from
    #  // uvm_tlm_generic_payload. Instead, you would use the extensions mechanism.
    #  // Thus in SystemC none of the accessors are virtual.
    #  //--------------------------------------------------------------------


    #  // Get the value of the <m_command> variable
    def get_command(self):
        return self.m_command


    #  // Set the value of the <m_command> variable
    def set_command(self, command):
        self.m_command = command


    #   // Function -- NODOCS -- is_read
    #   //
    #   // Returns true if the current value of the <m_command> variable
    #   // is ~UVM_TLM_READ_COMMAND~.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.15
    def is_read(self):
        return (self.m_command == UVM_TLM_READ_COMMAND)

    #   // Function -- NODOCS -- set_read
    #   //
    #   // Set the current value of the <m_command> variable
    #   // to ~UVM_TLM_READ_COMMAND~.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.16
    def set_read(self):
        self.set_command(UVM_TLM_READ_COMMAND)

    #   // Function -- NODOCS -- is_write
    #   //
    #   // Returns true if the current value of the <m_command> variable
    #   // is ~UVM_TLM_WRITE_COMMAND~.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.17
    def is_write(self):
        return (self.m_command == UVM_TLM_WRITE_COMMAND)

    #   // Function -- NODOCS -- set_write
    #   //
    #   // Set the current value of the <m_command> variable
    #   // to ~UVM_TLM_WRITE_COMMAND~.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.18
    def set_write(self):
        self.set_command(UVM_TLM_WRITE_COMMAND)


    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.20
    def set_address(self, addr):
        self.m_address = addr

    #   // Function -- NODOCS -- get_address
    #   //
    #   // Get the value of the <m_address> variable
    #
    def get_address(self):
        return self.m_address

    #   // Function -- NODOCS -- get_data
    #   //
    #   // Return the value of the <m_data> array
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.21
    #  virtual def void get_data (self,output byte unsigned p []):
    #    p = m_data
    #  endfunction

    #   // Function -- NODOCS -- set_data
    #   //
    #   // Set the value of the <m_data> array
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.22
    #  def set_data(self,ref byte unsigned p []):
    #    m_data = p
    #  endfunction

    #   // Function -- NODOCS -- get_data_length
    #   //
    #   // Return the current size of the <m_data> array
    #
    #  virtual def int unsigned get_data_length(self):
    #    return m_length
    #  endfunction

    #  // Function -- NODOCS -- set_data_length
    #  // Set the value of the <m_length>
    #
    #   // @uvm-ieee 1800.2-2017 auto 12.3.4.2.24
    #   def set_data_length(self,int unsigned length):
    #    m_length = length
    #  endfunction

    #   // Function -- NODOCS -- get_streaming_width
    #   //
    #   // Get the value of the <m_streaming_width> array
    #
    #  virtual def int unsigned get_streaming_width(self):
    #    return m_streaming_width
    #  endfunction


    #   // Function -- NODOCS -- set_streaming_width
    #   //
    #   // Set the value of the <m_streaming_width> array
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.26
    #  def set_streaming_width(self,int unsigned width):
    #    m_streaming_width = width
    #  endfunction


    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.27
    #  def get_byte_enable(self,output byte unsigned p[]):
    #    p = m_byte_enable
    #  endfunction
    #
    #   // Function -- NODOCS -- set_byte_enable
    #   //
    #   // Set the value of the <m_byte_enable> array
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.28
    #  def set_byte_enable(self,ref byte unsigned p[]):
    #    m_byte_enable = p
    #  endfunction
    #
    #   // Function -- NODOCS -- get_byte_enable_length
    #   //
    #   // Return the current size of the <m_byte_enable> array
    #
    #  virtual def int unsigned get_byte_enable_length(self):
    #    return m_byte_enable_length
    #  endfunction
    #
    #   // Function -- NODOCS -- set_byte_enable_length
    #   //
    #   // Set the size <m_byte_enable_length> of the <m_byte_enable> array
    #   // i.e.  <m_byte_enable>.size()
    #
    # // @uvm-ieee 1800.2-2017 auto 12.3.4.2.30
    # def set_byte_enable_length(self,int unsigned length):
    #    m_byte_enable_length = length
    #  endfunction
    #
    #   // Function -- NODOCS -- set_dmi_allowed
    #   //
    #   // DMI hint. Set the internal flag <m_dmi> to allow dmi access
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.31
    #  def set_dmi_allowed(self,bit dmi):
    #    m_dmi = dmi
    #  endfunction
    #
    #   // Function -- NODOCS -- is_dmi_allowed
    #   //
    #   // DMI hint. Query the internal flag <m_dmi> if allowed dmi access
    #
    # // @uvm-ieee 1800.2-2017 auto 12.3.4.2.32
    # def is_dmi_allowed(self):
    #    return m_dmi
    #  endfunction
    #
    #   // Function -- NODOCS -- get_response_status
    #   //
    #   // Return the current value of the <m_response_status> variable
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.33
    #  def get_response_status(self):
    #    return m_response_status
    #  endfunction
    #
    #   // Function -- NODOCS -- set_response_status
    #   //
    #   // Set the current value of the <m_response_status> variable
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.34
    #  def set_response_status(self,uvm_tlm_response_status_e status):
    #    m_response_status = status
    #  endfunction
    #
    #   // Function -- NODOCS -- is_response_ok
    #   //
    #   // Return TRUE if the current value of the <m_response_status> variable
    #   // is ~UVM_TLM_OK_RESPONSE~
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.35
    #  def is_response_ok(self):
    #    return (m_response_status) > 0  # cast to 'int' removed
    #  endfunction
    #
    #   // Function -- NODOCS -- is_response_error
    #   //
    #   // Return TRUE if the current value of the <m_response_status> variable
    #   // is not ~UVM_TLM_OK_RESPONSE~
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.36
    #  def is_response_error(self):
    #    return !is_response_ok()
    #  endfunction


    #   // Function -- NODOCS -- get_response_string
    #   //
    #   // Return the current value of the <m_response_status> variable
    #   // as a string
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.37
    #  def get_response_string(self):
    #
    #    case(m_response_status)
    #      UVM_TLM_OK_RESPONSE                : return "OK"
    #      UVM_TLM_INCOMPLETE_RESPONSE        : return "INCOMPLETE"
    #      UVM_TLM_GENERIC_ERROR_RESPONSE     : return "GENERIC_ERROR"
    #      UVM_TLM_ADDRESS_ERROR_RESPONSE     : return "ADDRESS_ERROR"
    #      UVM_TLM_COMMAND_ERROR_RESPONSE     : return "COMMAND_ERROR"
    #      UVM_TLM_BURST_ERROR_RESPONSE       : return "BURST_ERROR"
    #      UVM_TLM_BYTE_ENABLE_ERROR_RESPONSE : return "BYTE_ENABLE_ERROR"
    #    endcase
    #
    #    // we should never get here
    #    return "UNKNOWN_RESPONSE"
    #
    #  endfunction

    #  //--------------------------------------------------------------------
    #  // Group -- NODOCS -- Extensions Mechanism
    #  //
    #  //--------------------------------------------------------------------

    #  // Function -- NODOCS -- set_extension
    #  //
    #  // Add an instance-specific extension. Only one instance of any given
    #  // extension type is allowed. If there is an existing extension
    #  // instance of the type of ~ext~, ~ext~ replaces it and its handle
    #  // is returned. Otherwise, ~null~ is returned.
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.41
    #  def uvm_tlm_extension_base set_extension(self,uvm_tlm_extension_base ext):
    #    uvm_tlm_extension_base ext_handle = ext.get_type_handle()
    #    if(!m_extensions.exists(ext_handle))
    #      set_extension = None
    #    else
    #      set_extension = m_extensions[ext_handle]
    #    m_extensions[ext_handle] = ext
    #  endfunction


    #  // Function -- NODOCS -- get_num_extensions
    #  //
    #  // Return the current number of instance specific extensions.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.39
    #  def int get_num_extensions(self):
    #    return m_extensions.num()
    #  endfunction: get_num_extensions
    #
    #
    #  // Function -- NODOCS -- get_extension
    #  //
    #  // Return the instance specific extension bound under the specified key.
    #  // If no extension is bound under that key, ~null~ is returned.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.40
    #  def uvm_tlm_extension_base get_extension(self,uvm_tlm_extension_base ext_handle):
    #    if(!m_extensions.exists(ext_handle))
    #      return None
    #    return m_extensions[ext_handle]
    #  endfunction
    #

    #  // Function -- NODOCS -- clear_extension
    #  //
    #  // Remove the instance-specific extension bound under the specified key.
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.42
    #  def void clear_extension(self,uvm_tlm_extension_base ext_handle):
    #    if(m_extensions.exists(ext_handle))
    #      m_extensions.delete(ext_handle)
    #    else
    #      `uvm_info("GP_EXT", sv.sformatf("Unable to find extension to clear"), UVM_MEDIUM)
    #  endfunction
    #
    #

    #  // Function -- NODOCS -- clear_extensions
    #  //
    #  // Remove all instance-specific extensions
    #
    #  // @uvm-ieee 1800.2-2017 auto 12.3.4.2.43
    #  def void clear_extensions(self):
    #    m_extensions.delete()
    #  endfunction
    #
    #

    #  // Function -- NODOCS -- pre_randomize()
    #  // Prepare this class instance for randomization
    #  //
    #  def void pre_randomize(self):
    #    int i
    #    m_rand_exts = new [m_extensions.num()]
    #    foreach (m_extensions[ext_]):
    #      uvm_tlm_extension_base ext = ext_
    #      m_rand_exts[i++] = m_extensions[ext]
    #    end
    #  endfunction
    #

    #  // Function -- NODOCS -- post_randomize()
    #  // Clean-up this class instance after randomization
    #  //
    #  def void post_randomize(self):
    #     m_rand_exts.delete()
    #  endfunction
    #endclass


uvm_object_utils(UVMTLMGenericPayload)

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_gp
#//
#// This typedef provides a short, more convenient name for the
#// <uvm_tlm_generic_payload> type.
#//----------------------------------------------------------------------


#//----------------------------------------------------------------------
#// Class: uvm_tlm_extension_base
#//
#// The class uvm_tlm_extension_base is the non-parameterized base class for
#// all generic payload extensions.  It includes the utility do_copy()
#// and create().  The pure virtual function get_type_handle() allows you
#// to get a unique handle that represents the derived type.  This is
#// implemented in derived classes.
#//
#// This class is never used directly by users.
#// The <uvm_tlm_extension> class is used instead.
#//
class UVMTLMExtensionBase(UVMObject):

    #  // Function: new
    #  //
    def __init__(self, name = ""):
        super().__init__(name)

    #  // Function: get_type_handle
    #  //
    #  // An interface to polymorphically retrieve a handle that uniquely
    #  // identifies the type of the sub-class
    #
    #  pure virtual function uvm_tlm_extension_base get_type_handle();


    #  // Function: get_type_handle_name
    #  //
    #  // An interface to polymorphically retrieve the name that uniquely
    #  // identifies the type of the sub-class
    #
    #  pure virtual function string get_type_handle_name();


    def do_copy(self, rhs):
        super().do_copy(rhs)

    #  // Function: create
    def create(self, name=""):
        return None


#//----------------------------------------------------------------------
#// Class: uvm_tlm_extension
#//
#// TLM extension class. The class is parameterized with arbitrary type
#// which represents the type of the extension. An instance of the
#// generic payload can contain one extension object of each type; it
#// cannot contain two instances of the same extension type.
#//
#// The extension type can be identified using the <ID()>
#// method.
#//
#// To implement a generic payload extension, simply derive a new class
#// from this class and specify the name of the derived class as the
#// extension parameter.
#//
#//|
#//| class my_ID extends uvm_tlm_extension#(my_ID);
#//|   int ID;
#//|
#//|   `uvm_object_utils_begin(my_ID)
#//|      `uvm_field_int(ID, UVM_ALL_ON)
#//|   `uvm_object_utils_end
#//|
#//|   function new(string name = "my_ID");
#//|      super.new(name);
#//|   endfunction
#//| endclass
#//|
#
class UVMTLMExtension(UVMTLMExtensionBase):

    #   local static this_type m_my_tlm_ext_type = ID();

    #   // Function: new
    #   //
    #   // creates a new extension object.
    #
    def __init__(self, name=""):
        super().__init__(name)

    #   // Function: ID()
    #   //
    #   // Return the unique ID of this TLM extension type.
    #   // This method is used to identify the type of the extension to retrieve
    #   // from a <uvm_tlm_generic_payload> instance,
    #   // using the <uvm_tlm_generic_payload::get_extension()> method.
    #   //
    #  static function this_type ID();
    #    if (m_my_tlm_ext_type == null)
    #      m_my_tlm_ext_type = new();
    #    return m_my_tlm_ext_type;
    #  endfunction
    #
    #  virtual function uvm_tlm_extension_base get_type_handle();
    #     return ID();
    #  endfunction
    #
    #  virtual function string get_type_handle_name();
    #    return `uvm_typename(T);
    #  endfunction
    #
    #  virtual function uvm_object create (string name="");
    #    return null;
    #  endfunction



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


from ubus_transfer import *
from uvm.seq import UVMSequence
from uvm.base import sv, UVM_MEDIUM
from uvm.macros import *


DEBUG = True
def _print(msg):
    if DEBUG:
        print("[slave_mem_seq]: " + msg)

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: simple_response_seq
#//
#//------------------------------------------------------------------------------

#class simple_response_seq extends uvm_sequence #(ubus_transfer)
#   ubus_slave_sequencer p_sequencer
#
#  function new(string name="simple_response_seq")
#    super.new(name)
#  endfunction
#
#//  `uvm_object_utils(simple_response_seq)
#
#  `uvm_object_utils(simple_response_seq)
#
#  ubus_transfer util_transfer
#
#  virtual task body()
#     $cast(p_sequencer, m_sequencer)
#
#    `uvm_info(get_type_name(),
#      $sformatf("%s starting...",
#      get_sequence_path()), UVM_MEDIUM)
#    forever begin
#      p_sequencer.addr_ph_port.peek(util_transfer)
#
#      // Need to raise/drop objection before each item because we don't want
#      // to be stopped in the middle of a transfer.
#      uvm_test_done.raise_objection(this)
#      `uvm_do_with(req,
#        { req.read_write == util_transfer.read_write
#          req.size == util_transfer.size
#          req.error_pos == 1000; } )
#      uvm_test_done.drop_objection(this)
#    end
#  endtask : body
#
#endclass : simple_response_seq
#
#

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: slave_memory_seq
#//
#//------------------------------------------------------------------------------


#class slave_memory_seq extends uvm_sequence #(ubus_transfer)
class slave_memory_seq(UVMSequence):
    #
    def __init__(self, name="slave_memory_seq"):
        UVMSequence.__init__(self, name)
        #  ubus_transfer util_transfer
        self.util_transfer = ubus_transfer()
        self.m_mem = {}
        self.req = ubus_transfer()

    #  `uvm_declare_p_sequencer(ubus_slave_sequencer)

    def pre_do(self, is_item):
        # Update the properties that are relevant to both read and write
        _print("calling PRE_DO now: " + self.util_transfer.convert2string())
        self.req.size       = self.util_transfer.size
        self.req.addr       = self.util_transfer.addr
        self.req.read_write = self.util_transfer.read_write
        self.req.error_pos  = 1000
        self.req.transmit_delay = 0
        self.req.data = [0] * self.util_transfer.size
        self.req.wait_state = [0] * self.util_transfer.size
        for i in range(self.util_transfer.size):
            self.req.wait_state[i] = 2
            # For reads, populate req with the random "readback" data of the size
            # requested in util_transfer
            if (self.req.read_write == READ):
                new_addr = self.util_transfer.addr + i
                if new_addr not in self.m_mem:
                    self.m_mem[new_addr] = sv.urandom()
                self.req.data[i] = self.m_mem[new_addr]


    def post_do(self, this_item):
        # For writes, update the m_mem associative array
        if self.util_transfer.read_write == WRITE:
            for i in range(self.req.size):
                self.m_mem[self.req.addr + i] = self.req.data[i]


    async def body(self):
        #p = None  # uvm_phase
        uvm_info(self.get_name(), sv.sformatf("ubus_slave_seq %s starting...",
            self.get_sequence_path()), UVM_MEDIUM)

        #$cast(req, create_item(ubus_transfer::get_type(), p_sequencer, "req"))
        #p = self.get_starting_phase()
        self.req = self.create_item(ubus_transfer.get_type(), self.p_sequencer,
            "req")

        while True:
            util_transfer = []
            await self.p_sequencer.addr_ph_port.peek(util_transfer)
            self.util_transfer = util_transfer[0]
            _print("slave_mem_seq after peek. Got util_xfer: " +
                self.util_transfer.convert2string())

            # Need to raise/drop objection before each item because we don't want
            # to be stopped in the middle of a transfer.
            # p.raise_objection(self)

            _print("BEFORE start_item. req is " +
                self.req.convert2string())
            await self.start_item(self.req)
            _print("after start_item")
            await self.finish_item(self.req)
            _print("after finish_item")
            #p.drop_objection(self)


uvm_object_utils(slave_memory_seq)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from ..base.uvm_object import UVMObject

#// Class: uvm_set_get_dap_base
#// Provides the 'set' and 'get' interface for Data Access Policies (DAPs)
#//
#// The 'Set/Get' base class simply provides a common interface for
#// the various DAPs to implement.  This provides a mechanism for
#// consistent implementations of similar DAPs.
#//
#virtual class uvm_set_get_dap_base#(type T=int) extends uvm_object;


class uvm_set_get_dap_base(UVMObject):


    def __init__(self, name="unnamed-uvm_set_get_dap_base#(T)"):
        UVMObject.__init__(self, name)

    #   // Group: Set/Get Interface
    #   //
    #   // All implementations of the ~uvm_set_get_dap_base~ class must
    #   // provide an implementation of the four basic "Set and Get"
    #   // accessors.
    #   //

    #   // Function: set
    #   // Sets the value contained within the resource.
    #   //
    #   // Depending on the DAP policies, an error may be reported if
    #   // it is illegal to 'set' the value at this time.
    #   pure virtual function void set(T value);
    def set(self, value):
        raise NotImplementedError('Pure virtual function')

    #   // Function: try_set
    #   // Attempts to set the value contained within the resource.
    #   //
    #   // If the DAP policies forbid setting at this time, then
    #   // the method will return 0, however no errors will be
    #   // reported.  Otherwise, the method will return 1, and
    #   // will be treated like a standard <set> call.
    #   pure virtual function bit try_set(T value);
    def try_set(self, value):
        raise NotImplementedError('Pure virtual function')

    #   // Function: get
    #   // Retrieves the value contained within the resource.
    #   //
    #   // Depending on the DAP policies, an error may be reported
    #   // if it is illegal to 'get' the value at this time.
    #   pure virtual function T get();
    def get(self):
        raise NotImplementedError('Pure virtual function')

    #   // Function: try_get
    #   // Attempts to retrieve the value contained within the resource.
    #   //
    #   // If the DAP policies forbid retrieving at this time, then
    #   // the method will return 0, however no errors will be
    #   // reported.  Otherwise, the method will return 1, and will
    #   // be treated like a standard <get> call.
    #   pure virtual function bit try_get(output T value);
    def try_get(self, value):
        raise NotImplementedError('Pure virtual function')



#//
#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2010-2011 Cadence Design Systems, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
from uvm.comps.uvm_test import UVMTest
from uvm.macros import *
import cocotb
from blk_seqlib import *
from uvm.base.sv import sv
from uvm.base.uvm_root import uvm_top
from blk_env import blk_env

from blk_seqlib import *

class blk_R_test(UVMTest):

    #
    #   blk_env env
    #
    def __init__(self, name="blk_R_test", parent=None):
        super().__init__(name, parent)
        self.env = None  # blk_env

    def build_phase(self, phase):
        if (self.env is None):
            arr = []
            if (sv.cast(arr, uvm_top.find("env$"), blk_env)):
                self.env = arr[0]
            else:
                uvm_fatal("NO_ENV_ERR", "test_top unable to find 'env'")

    
    async def run_phase(self, phase):
        reset_seq = None  # uvm_sequence_base
        seq = None  # blk_R_test_seq
        phase.raise_objection(self)

        rst_seq = dut_reset_seq.type_id.create("rst_seq", self)
        rst_seq.vif = self.env.apb.vif
        await rst_seq.start(None)

        self.env.model.reset()

        seq = blk_R_test_seq.type_id.create("blk_R_test_seq",self)
        seq.model = self.env.model
        await seq.start(None)

        phase.drop_objection(self)


uvm_component_utils(blk_R_test)
